from .ase import ASEIO, ASECreate
from .benchmark import benchmark_read
from .chemfiles import ChemfilesIO
from .generate import create_frames
from .mda import MDAIO
from .mdtraj import MDTrajIO
from .plams import PLAMSIO
from .znh5md import ZnH5MDFixedShapeIO, ZnH5MDIO

__all__ = [
    "create_frames",
    "ASEIO",
    "ASECreate",
    "benchmark_read",
    "MDAIO",
    "ChemfilesIO",
    "MDTrajIO",
    "PLAMSIO",
    "ZnH5MDIO",
    "ZnH5MDFixedShapeIO",
]



import typing as t
from enum import Enum

if t.TYPE_CHECKING:
    from znh5md.serialization import Frames


class AttributePath(str, Enum):
    origin = "ASE_ENTRY_ORIGIN"
    unit = "unit"
    boundary = "boundary"
    dimension = "dimension"


class H5MDPath(str, Enum):
    # https://www.nongnu.org/h5md/h5md.html#particles-group
    position = "/particles/{}/position"
    velocity = "/particles/{}/velocity"
    force = "/particles/{}/force"
    mass = "/particles/{}/mass"
    species = "/particles/{}/species"
    id = "/particles/{}/id"
    charge = "/particles/{}/charge"

    # https://www.nongnu.org/h5md/modules/thermodynamics.html
    pressure = "/observables/{}/pressure"
    temperature = "/observables/{}/temperature"
    density = "/observables/{}/density"
    potential_energy = "/observables/{}/potential_energy"
    kinetic_energy = "/observables/{}/kinetic_energy"
    internal_energy = "/observables/{}/internal_energy"
    enthalpy = "/observables/{}/enthalpy"

    # cell
    box_edges = "/particles/{}/box/edges"
    box_pbc = "/particles/{}/box/pbc"


class H5MDToASEMapping(str, Enum):
    position = "positions"
    force = "forces"
    mass = "masses"
    charge = "charges"
    potential_energy = "energy"
    box_edges = "cell"
    box_pbc = "pbc"
    species = "numbers"
    velocity = "velocities"


def get_h5md_path(name: str, particles_group: str, frames: "Frames") -> str:
    try:
        h5md_name = H5MDToASEMapping(name).name
    except ValueError:
        h5md_name = name
    try:
        return H5MDPath[h5md_name].format(particles_group)
    except KeyError:
        if name in frames.arrays.keys():
            return "/particles/{}/{}".format(particles_group, name)
        elif name in frames.info.keys():
            return "/observables/{}/{}".format(particles_group, name)
        elif name in frames.calc.keys():
            in_particles = False
            try:
                in_particles = (
                    frames.calc[name][0].shape[0] == frames.positions[0].shape[0]
                )
                # it could be a coincidence that the shape of the property is the same
                #  as the shape of the positions
                # e.g for 3, or 6 particles!
            except (AttributeError, IndexError):
                # e.g. object has no attribute 'shape'
                # or tuple index out of range
                pass

            if in_particles:
                return "/particles/{}/{}".format(particles_group, name)
            else:
                return "/observables/{}/{}".format(particles_group, name)
        else:
            raise ValueError(f"Unable to determine path for '{name}'")



import contextlib
import dataclasses
import importlib.metadata
import os
import pathlib
import typing as t
import warnings
from collections.abc import MutableSequence

import ase
import h5py
import numpy as np

from znh5md.interface.read import getitem
from znh5md.interface.write import extend
from znh5md.misc import open_file

__version__ = importlib.metadata.version("znh5md")


@dataclasses.dataclass
class IO(MutableSequence):
    """
    A class for handling H5MD files for ASE Atoms objects.

    This class provides an interface to read from and write to H5MD files,
    treating them as a sequence of ASE Atoms objects. It supports various
    configurations for file handling, data compression, and metadata.

    Parameters
    ----------
    filename : str | os.PathLike | None, optional
        Path to the H5MD file. If `file_handle` is not provided, this must be specified.
        Defaults to None.
    file_handle : h5py.File | None, optional
        An existing h5py file handle. If `filename` is not provided, this must
        be specified.
        Defaults to None.
    file_factory : Callable[[], ContextManager[h5py.File]] | None, optional
        A factory function that returns a context manager for an h5py file.
        File factory is restricted to read-only data access.
    pbc_group : bool, optional
        Specify if Periodic Boundary Conditions (PBC) should be stored per step.
        This is not H5MD conformant, but it allows for more flexibility in data storage.
        Defaults to True.
    save_units : bool, optional
        Export ASE units into the H5MD file. Defaults to True.
    author : str, optional
        Author's name for H5MD metadata. Defaults to "N/A".
    author_email : str, optional
        Author's email for H5MD metadata. Defaults to "N/A".
    creator : str, optional
        Name of the creating software. Defaults to "znh5md".
    creator_version : str, optional
        Version of the creating software. Defaults to the znh5md package version.
    particles_group : str | None, optional
        Name of the particles group within the H5MD file (e.g., "atoms").
        If None, it tries to infer from an existing file or defaults to "atoms".
        Defaults to None.
    compression : str | None, optional
        Compression filter to use for datasets (e.g., "gzip", "lzf").
        Defaults to "gzip".
    compression_opts : int | None, optional
        Compression options (e.g., compression level for gzip). Defaults to None.
    timestep : float, optional
        Time step between frames in the H5MD file. Defaults to 1.0.
    store : t.Literal["time", "linear"], optional
        Method for storing time and step information.
        "time" stores a time series, "linear" stores a single value.
        Defaults to "linear".
    tqdm_limit : int, optional
        Threshold for using tqdm progress bar. Defaults to 100.
    chunk_size : int | None | list[int] | tuple[int, ...], optional
        Chunk shape for HDF5 datasets. Can be an integer, list, or tuple.
        Defaults to (64, 64).
    use_ase_calc : bool, optional
        Whether to use ASE calculator for storing and retrieving data.
        Defaults to True.
    variable_shape : bool, optional
        Whether the data has a variable shape across frames. Defaults to True.
        This is not H5MD conformant, but allows for more flexibility in data storage.
    include : list[str] | None, optional
        List of attributes to include when reading/writing. If specified,
        "position" must be included. Defaults to None, which means all
        data is included.
    mask : list[int] | slice | None, optional
        Mask to apply when reading data, e.g., to select specific atoms.
        Not supported with `variable_shape=True`. Defaults to None.

    Raises
    ------
    ValueError
        If both `filename` and `file_handle` are provided or neither is provided.
        If "position" is not in `include` list when `include` is specified.
        If `mask` is used with `variable_shape=True`.

    Examples
    --------
    >>> import numpy as np
    >>> from ase.build import bulk
    >>> from znh5md import IO
    >>>
    >>> # Create some dummy atoms objects
    >>> atoms1 = bulk("Cu", "fcc", a=3.6)
    >>> atoms2 = atoms1.copy()
    >>> atoms2.positions[0, 0] += 0.1
    >>>
    >>> # Write to an H5MD file
    >>> io = IO(filename="test.h5")
    >>> io.extend([atoms1, atoms2])
    >>>
    >>> # Read from the H5MD file
    >>> io_read = IO(filename="test.h5")
    >>> print(len(io_read))
    2
    >>> atoms_read = io_read[0]
    >>> print(atoms_read.positions[0])
    [0. 0. 0.]
    >>>
    >>> # Clean up
    >>> import os
    >>> os.remove("test.h5")
    """

    filename: str | os.PathLike | None = None
    file_handle: h5py.File | None = None
    file_factory: t.Callable[[], t.ContextManager[h5py.File]] | None = None
    pbc_group: bool = True
    save_units: bool = True
    author: str = "N/A"
    author_email: str = "N/A"
    creator: str = "znh5md"
    creator_version: str = __version__
    particles_group: str | None = None
    compression: str | None = "gzip"
    compression_opts: int | None = None
    timestep: float = 1.0
    store: t.Literal["time", "linear"] = "linear"
    tqdm_limit: int = 100
    chunk_size: int | None | list[int] | tuple[int, ...] = (64, 64)
    use_ase_calc: bool = True
    variable_shape: bool = True
    include: list[str] | None = None
    mask: list[int] | slice | None = None

    _store_ase_origin: bool = True  # for testing purposes only

    def __post_init__(self):
        sources = [
            self.filename is not None,
            self.file_handle is not None,
            self.file_factory is not None,
        ]
        if sum(sources) == 0:
            raise ValueError(
                "Either filename, file_handle, or file_factory must be provided"
            )
        if sum(sources) > 1:
            raise ValueError(
                "Only one of filename, file_handle, or file_factory can be provided"
            )
        if self.filename is not None:
            self.filename = pathlib.Path(self.filename)
        self._set_particle_group()
        self._read_author_creator()
        if self.include is not None:
            if "position" not in self.include:
                raise ValueError("'position' must be in keys")
        if self.mask is not None and self.variable_shape is True:
            raise ValueError("mask is not supported with variable_shape=True. ")

    def _set_particle_group(self):
        if self.particles_group is not None:
            pass
        elif self.filename is not None and self.filename.exists():
            with open_file(
                self.filename, self.file_handle, self.file_factory, mode="r"
            ) as f:
                self.particles_group = next(iter(f["particles"].keys()))
        elif (
            self.file_handle is not None
            and pathlib.Path(self.file_handle.filename).exists()
        ):
            with open_file(
                self.filename, self.file_handle, self.file_factory, mode="r"
            ) as f:
                self.particles_group = next(iter(f["particles"].keys()))
        else:
            self.particles_group = "atoms"  # Default group name

    def _read_author_creator(self):
        with contextlib.suppress(FileNotFoundError, KeyError):
            # FileNotFoundError if the filename does not exist
            # KeyError if the file has not yet been initialized as H5MD
            #   or the keys are not provided, which is officially
            #   not allowed in H5MD.
            with open_file(
                self.filename, self.file_handle, self.file_factory, mode="r"
            ) as f:
                self.author = f["h5md"]["author"].attrs["name"]
                self.author_email = f["h5md"]["author"].attrs["email"]
                self.creator = f["h5md"]["creator"].attrs["name"]
                self.creator_version = f["h5md"]["creator"].attrs["version"]

    def create_file(self):
        with open_file(
            self.filename, self.file_handle, self.file_factory, mode="w"
        ) as f:
            g_h5md = f.create_group("h5md")
            g_h5md.attrs["version"] = np.array([1, 1])
            g_author = g_h5md.create_group("author")
            g_author.attrs["name"] = self.author
            g_author.attrs["email"] = self.author_email
            g_creator = g_h5md.create_group("creator")
            g_creator.attrs["name"] = self.creator
            g_creator.attrs["version"] = self.creator_version
            f.create_group("particles")

    def __len__(self) -> int:
        with open_file(
            self.filename, self.file_handle, self.file_factory, mode="r"
        ) as f:
            return len(f["particles"][self.particles_group]["species"]["value"])

    @t.overload
    def __getitem__(self, index: int) -> ase.Atoms: ...
    @t.overload
    def __getitem__(self, index: np.int_) -> ase.Atoms: ...
    @t.overload
    def __getitem__(self, index: slice) -> list[ase.Atoms]: ...
    @t.overload
    def __getitem__(self, index: np.ndarray) -> list[ase.Atoms]: ...
    @t.overload
    def __getitem__(self, index: list[int]) -> list[ase.Atoms]: ...

    def __getitem__(
        self, index: int | np.int_ | slice | np.ndarray | list[int]
    ) -> ase.Atoms | list[ase.Atoms]:
        try:
            return getitem(self, index)
        except FileNotFoundError:
            # FileNotFoundError is an OSError, but we want to handle it
            #  separately from the OSError h5py raises
            raise
        except OSError:
            raise IndexError("Index out of range")

    def extend(self, frames: list[ase.Atoms]) -> None:
        if not isinstance(frames, list):
            raise ValueError("images must be a list of ASE Atoms objects")
        if len(frames) == 0:
            warnings.warn("No data provided")
            return
        if self.filename is not None and not self.filename.exists():
            self.create_file()
        if self.file_handle is not None:
            needs_creation = False
            with open_file(
                self.filename, self.file_handle, self.file_factory, mode="r"
            ) as f:
                needs_creation = "h5md" not in f
            if needs_creation:
                self.create_file()

        extend(self, frames)

    def append(self, atoms: ase.Atoms):
        if not isinstance(atoms, ase.Atoms):
            raise ValueError("atoms must be an ASE Atoms object")
        self.extend([atoms])

    def __delitem__(self, index):
        raise NotImplementedError("Deleting items is not supported")

    def __setitem__(self, index, value):
        raise NotImplementedError("Setting items is not supported")

    def insert(self, index, value):
        raise NotImplementedError("Inserting items is not supported")



import numpy.testing as npt
from ase.build import molecule
from ase.calculators.singlepoint import SinglePointCalculator

import znh5md
import znh5md.serialization


def test_single_entry_info(tmp_path):
    # Test a special case where only the first config has the key
    # which caused an error in the past
    io = znh5md.IO(tmp_path / "test.h5")
    water = molecule("H2O")
    water.info["density"] = 0.997
    io.append(water)
    del water.info["density"]
    io.extend([water for _ in range(5)])
    assert len(io) == 6
    assert len(list(io)) == 6
    assert len(io[:]) == 6
    assert io[0].info["density"] == 0.997
    assert "density" not in io[1].info

    frames = znh5md.serialization.Frames.from_ase(list(io))
    assert len(frames) == 6
    assert len(list(frames)) == 6


def test_single_entry_arrays(tmp_path):
    # Test a special case where only the first config has the key
    # which caused an error in the past
    io = znh5md.IO(tmp_path / "test.h5")
    water = molecule("H2O")
    water.arrays["density"] = [0.997, 0.998, 0.999]
    io.append(water)
    del water.arrays["density"]
    io.extend([water for _ in range(5)])
    assert len(io) == 6
    assert len(list(io)) == 6
    assert len(io[:]) == 6
    npt.assert_array_equal(io[0].arrays["density"], [0.997, 0.998, 0.999])
    assert "density" not in io[1].arrays

    frames = znh5md.serialization.Frames.from_ase(list(io))
    assert len(frames) == 6
    assert len(list(frames)) == 6


def test_single_entry_calc(tmp_path):
    # Test a special case where only the first config has the key
    # which caused an error in the past
    io = znh5md.IO(tmp_path / "test.h5")
    water = molecule("H2O")
    water.calc = SinglePointCalculator(water, energy=0.0, forces=[0.0, 0.0, 0.0])
    io.append(water)
    water.calc = None
    io.extend([water for _ in range(5)])
    assert len(io) == 6
    assert len(list(io)) == 6
    assert len(io[:]) == 6
    assert io[0].calc.results["energy"] == 0.0
    assert io[1].calc is None

    frames = znh5md.serialization.Frames.from_ase(list(io))
    assert len(frames) == 6
    assert len(list(frames)) == 6



import numpy.testing as npt

import znh5md


def test_variable_shape_off(tmp_path, full_water):
    """Test the variable shape off functionality."""
    io = znh5md.IO(tmp_path / "test.h5", variable_shape=False)
    io.extend([full_water for _ in range(10)])

    assert len(io) == 10
    assert len(list(io)) == 10
    for atoms in io:
        assert atoms == full_water
        for key, val in atoms.arrays.items():
            npt.assert_equal(
                val,
                full_water.arrays[key],
            )
        for key, val in atoms.info.items():
            npt.assert_equal(
                val,
                full_water.info[key],
            )
        for key, val in atoms.calc.results.items():
            npt.assert_equal(
                val,
                full_water.calc.results[key],
            )



import numpy.testing as npt

import znh5md


def test_info_non_ascii(tmp_path, frames_with_residuenames):
    io = znh5md.IO(tmp_path / "test.h5")
    # io = znh5md.IO("test.h5")
    io.extend(frames_with_residuenames)

    for a, b in zip(io, frames_with_residuenames):
        for key in b.arrays:
            npt.assert_array_equal(a.arrays[key], b.arrays[key])
            # test json per atom?



import uuid

import ase
import numpy as np
import pytest

import znh5md


@pytest.fixture
def atoms():
    def _create_atoms(count: int, size: int) -> list[ase.Atoms]:
        """Create random atoms.

        Attributes
        ----------
        count : int
            Number of images.
        size : int
            Number of atoms in each image.
        """
        images = []
        for _ in range(count):
            images.append(ase.Atoms("H" * size, positions=np.random.rand(size, 3)))
        return images

    return _create_atoms


@pytest.mark.benchmark(group="write-count")
@pytest.mark.parametrize("count", [100 * x for x in range(1, 11)])
def test_bm_write_count(tmp_path, benchmark, atoms, count, size=100):
    images = atoms(count, size)

    def _write():
        filename = tmp_path / f"{uuid.uuid4()}.h5md"
        znh5md.write(filename, images, compression=None)

    benchmark(_write)


@pytest.mark.benchmark(group="write-size")
@pytest.mark.parametrize("size", [100 * x for x in range(1, 11)])
def test_bm_write_size(tmp_path, benchmark, atoms, size, count=100):
    images = atoms(count, size)

    def _write():
        filename = tmp_path / f"{uuid.uuid4()}.h5md"
        znh5md.write(filename, images, compression=None)

    benchmark(_write)


@pytest.mark.benchmark(group="read-count")
@pytest.mark.parametrize("count", [100 * x for x in range(1, 11)])
def test_bm_read_count(tmp_path, benchmark, atoms, count, size=100):
    images = atoms(count, size)
    filename = tmp_path / f"{uuid.uuid4()}.h5md"
    znh5md.write(filename, images, compression="gzip")

    def _read():
        io = znh5md.IO(filename)
        _ = io[:]

    benchmark(_read)


@pytest.mark.benchmark(group="read-size")
@pytest.mark.parametrize("size", [100 * x for x in range(1, 11)])
def test_bm_read_size(tmp_path, benchmark, atoms, size, count=100):
    images = atoms(count, size)
    filename = tmp_path / f"{uuid.uuid4()}.h5md"
    znh5md.write(filename, images, compression="gzip")

    def _read():
        io = znh5md.IO(filename)
        _ = io[:]

    benchmark(_read)



import znh5md


def test_info_non_ascii(tmp_path, s22_no_ascii):
    io = znh5md.IO(tmp_path / "test.h5")
    io.extend(s22_no_ascii)

    for a, b in zip(io, s22_no_ascii):
        assert a.info["config"] == b.info["config"] == "βγ"



import contextlib
import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from src import (
    ASEIO,
    MDAIO,
    PLAMSIO,
    ASECreate,
    ChemfilesIO,
    MDTrajIO,
    ZnH5MDFixedShapeIO,
    ZnH5MDIO,
    benchmark_read,
    create_frames,
)
from tqdm import tqdm

IO_CLASSES = [
    ASEIO,
    MDTrajIO,
    MDAIO,
    ChemfilesIO,
    PLAMSIO,
    ASECreate,
    ZnH5MDIO,
    ZnH5MDFixedShapeIO,
]


def benchmark_io_for_frame_count(
    num_frames: int,
    num_atoms: int,
    file_format: str,
    filename: str = "test_io_benchmark",
):
    frames = create_frames(num_frames=num_frames, num_atoms=num_atoms)
    results = {}

    # Determine the actual format and compression based on the input file_format
    actual_format = file_format
    if file_format == "h5md-uncompressed":
        actual_format = "h5md"

    # Create the appropriate writer instance
    writer_instance = None
    if actual_format == "xtc":
        writer_instance = MDTrajIO(
            filename=filename,
            format=actual_format,
            num_atoms=num_atoms,
            num_frames=num_frames,
        )
    elif actual_format == "h5md":
        writer_instance = ZnH5MDIO(
            filename=filename,
            format=actual_format,
            num_atoms=num_atoms,
            num_frames=num_frames,
            compression=None if file_format == "h5md-uncompressed" else "gzip",
        )
    else:
        writer_instance = ASEIO(
            filename=filename,
            format=actual_format,
            num_atoms=num_atoms,
            num_frames=num_frames,
        )

    writer_instance.setup()
    writer_instance.write(frames)

    try:
        # Benchmark all reader classes
        for io_cls in IO_CLASSES:
            # Create reader instance, handling h5md compression for ZnH5MDIO
            if io_cls == ZnH5MDIO:
                io_obj = io_cls(
                    filename=filename,
                    format=actual_format,
                    num_atoms=num_atoms,
                    num_frames=num_frames,
                )
            elif io_cls == ZnH5MDFixedShapeIO:
                io_obj = io_cls(
                    filename=filename,
                    format=actual_format,
                    num_atoms=num_atoms,
                    num_frames=num_frames,
                )
            else:
                io_obj = io_cls(
                    filename=filename,
                    format=actual_format,
                    num_atoms=num_atoms,
                    num_frames=num_frames,
                )

            io_obj.setup()
            with contextlib.suppress(
                ValueError
            ):  # Catch ValueError for unsupported formats
                metrics = benchmark_read(io_obj)
                results[io_cls.__name__] = metrics.asdict()
    finally:
        if os.path.exists(filename):
            os.unlink(filename)
    return results


def compute_benchmark_dfs(full_results, num_atoms: int):
    df_mean = pd.DataFrame.from_dict(
        {
            k: {
                name: values["mean"] - v[ASECreate.__name__]["mean"]
                for name, values in v.items()
            }
            for k, v in full_results.items()
        },
        orient="index",
    )
    df_mean.index.name = "Number of Frames"

    df_std = pd.DataFrame.from_dict(
        {
            k: {name: values["std"] for name, values in v.items()}
            for k, v in full_results.items()
        },
        orient="index",
    )
    # convert to mus / atom
    df_avg = df_mean.divide(df_mean.index, axis=0).divide(num_atoms) * 1e6
    df_std_avg = df_std.divide(df_std.index, axis=0).divide(num_atoms) * 1e6

    return df_avg, df_std_avg


def plot_benchmarks(df_avg, df_std_avg, format: str):
    fig, ax = plt.subplots(figsize=(10, 6))
    colors = ["#1f77b4", "#ff7f0e", "#2ca02c", "#d62728", "#9467bd", "#8c564b"]
    markers = ["o", "s", "^", "v", "p", "*"]

    for i, library_name in enumerate(df_avg.columns):
        if library_name == ASECreate.__name__:
            continue
        color = colors[i % len(colors)]
        marker = markers[i % len(markers)]
        ax.plot(
            df_avg.index,
            df_avg[library_name],
            label=library_name,
            marker=marker,
            markersize=8,
            color=color,
            linewidth=2,
        )
        ax.errorbar(
            df_avg.index,
            df_avg[library_name],
            yerr=df_std_avg[library_name],
            fmt="none",
            capsize=5,
            color=color,
            alpha=0.7,
        )

    ax.set_xscale("log")
    ax.set_xlabel("Number of Frames")
    ax.set_ylabel(r"Mean runtime per atom / $\mu$s/atom")
    ax.set_title(f"Read Benchmark ({format.upper()} format)")
    ax.set_yscale("log")
    ax.legend(title="Library")
    ax.grid(which="both", linestyle="--", linewidth=0.5)
    plt.tight_layout()
    ax.set_ylim(bottom=1e-3, top=1e2)
    plt.savefig(f"benchmark_{format}.png", dpi=300)


def main():
    num_atoms = 512
    for format in ["xtc", "h5md", "xyz", "pdb", "h5md-uncompressed"]:
        print(f"Running benchmark for {format.upper()} format")
        full_results = {
            num_frames: benchmark_io_for_frame_count(
                num_atoms=num_atoms, num_frames=num_frames, file_format=format
            )
            for num_frames in tqdm(np.logspace(2, 3, num=10, dtype=int))
        }

        df_avg, df_std_avg = compute_benchmark_dfs(full_results, num_atoms=num_atoms)
        plot_benchmarks(df_avg, df_std_avg, format=format)


if __name__ == "__main__":
    main()



import h5py
import rdkit2ase

import znh5md


def test_extend_same_size(tmp_path):
    water = rdkit2ase.smiles2atoms("O")
    io = znh5md.IO(tmp_path / "test.h5")

    io.append(water)
    assert len(io) == 1

    with h5py.File(io.filename, "r") as f:
        assert f["particles/atoms/position/value"].shape == (1, 3, 3)

    io.append(water)
    assert len(io) == 2

    with h5py.File(io.filename, "r") as f:
        assert f["particles/atoms/position/value"].shape == (2, 3, 3)


def test_extend_different_size(tmp_path):
    water = rdkit2ase.smiles2atoms("O")
    ammonia = rdkit2ase.smiles2atoms("N")

    io = znh5md.IO(tmp_path / "test.h5")

    io.append(water)
    assert len(io) == 1

    with h5py.File(io.filename, "r") as f:
        assert f["particles/atoms/position/value"].shape == (1, 3, 3)

    io.append(ammonia)
    assert len(io) == 2

    with h5py.File(io.filename, "r") as f:
        assert f["particles/atoms/position/value"].shape == (2, 4, 3)


def test_extend_misssing(tmp_path):
    water = rdkit2ase.smiles2atoms("O")
    ammonia = rdkit2ase.smiles2atoms("N")

    ammonia.info["val"] = 1
    ammonia.arrays["mass"] = ammonia.get_masses()

    io = znh5md.IO(tmp_path / "test.h5")

    io.append(water)
    assert len(io) == 1

    with h5py.File(io.filename, "r") as f:
        assert f["particles/atoms/position/value"].shape == (1, 3, 3)
        assert "observables/atoms/val/value" not in f
        assert "particles/atoms/mass/value" not in f

    io.append(ammonia)
    assert len(io) == 2

    with h5py.File(io.filename, "r") as f:
        assert f["particles/atoms/position/value"].shape == (2, 4, 3)
        assert f["observables/atoms/val/value"].shape == (2,)
        assert f["particles/atoms/mass/value"].shape == (2, 4)

    # io.append(water)

    # with h5py.File(io.filename, "r") as f:
    #     assert f["particles/atoms/position/value"].shape == (3, 4, 3)
    #     assert f["observables/atoms/val/value"].shape == (2,)
    #     assert f["particles/atoms/mass/value"].shape == (2, 4)



import numpy as np
import pytest

import znh5md


@pytest.mark.parametrize(
    "dataset_name",
    [
        "s22",
        "s22_energy",
        "s22_energy_forces",
        "s22_all_properties",
        "s22_info_arrays_calc",
        "s22_mixed_pbc_cell",
        "s22_illegal_calc_results",
        "s22_no_ascii",
        "frames_with_residuenames",
        "s22_info_arrays_calc_missing_inbetween",
    ],
)
@pytest.mark.parametrize("append", [False])
def test_frames_iter(dataset_name, append, request, tmp_path):
    dataset = request.getfixturevalue(dataset_name)

    io = znh5md.IO(tmp_path / "test.h5")
    if append:
        for frame in dataset:
            io.append(frame)
    else:
        io.extend(dataset)

    for atoms in io:
        for value in atoms.arrays.values():
            assert isinstance(value, np.ndarray), f"Exp. np.ndarray, got {type(value)}"

    for atoms in io[:]:
        for value in atoms.arrays.values():
            assert isinstance(value, np.ndarray), f"Exp. np.ndarray, got {type(value)}"



import numpy as np
import numpy.testing as npt
import pytest

import znh5md


@pytest.mark.parametrize(
    "dataset",
    [
        "s22_info_arrays_calc",
    ],
)
def test_restricted_read_position(tmp_path, dataset, request):
    frames = request.getfixturevalue(dataset)
    io = znh5md.IO(tmp_path / "test.h5")
    io.extend(frames)

    rio = znh5md.IO(tmp_path / "test.h5", include=["position"])
    assert len(rio) == len(rio[:])
    assert len(rio) == len(frames)

    for atoms, ref in zip(rio, frames):
        assert atoms.calc is None
        assert atoms.info == {}
        # 'numbers' is always implicitly added when reading,
        #  even if only 'positions' is requested.
        # This is because 'numbers' defines the structure of the atom.
        assert atoms.arrays.keys() == {"numbers", "positions"}
        npt.assert_array_equal(atoms.get_atomic_numbers(), ref.get_atomic_numbers())
        npt.assert_array_equal(atoms.get_positions(), ref.get_positions())


def test_restricted_read_empty(tmp_path):
    with pytest.raises(ValueError):
        znh5md.IO(tmp_path / "test.h5", include=[])


@pytest.mark.parametrize(
    "dataset",
    [
        "s22_mixed_pbc_cell",
    ],
)
def test_restricted_read_box(tmp_path, dataset, request):
    frames = request.getfixturevalue(dataset)
    io = znh5md.IO(tmp_path / "test.h5")
    io.extend(frames)

    rio = znh5md.IO(tmp_path / "test.h5", include=["position", "box"])
    assert len(rio) == len(rio[:])
    assert len(rio) == len(frames)

    for atoms, ref in zip(rio, frames):
        assert atoms.calc is None
        assert atoms.info == {}
        assert atoms.arrays.keys() == {"numbers", "positions"}
        npt.assert_array_equal(atoms.get_atomic_numbers(), ref.get_atomic_numbers())
        npt.assert_array_equal(atoms.get_positions(), ref.get_positions())
        npt.assert_array_equal(atoms.get_cell(), ref.get_cell())
        npt.assert_array_equal(atoms.get_pbc(), ref.get_pbc())


@pytest.mark.parametrize(
    "dataset",
    [
        "s22_info_arrays_calc",
    ],
)
def test_restricted_read_potential_energy(tmp_path, dataset, request):
    frames = request.getfixturevalue(dataset)
    io = znh5md.IO(tmp_path / "test.h5")
    io.extend(frames)

    rio = znh5md.IO(tmp_path / "test.h5", include=["position", "potential_energy"])
    assert len(rio) == len(rio[:])
    assert len(rio) == len(frames)

    for atoms, ref in zip(rio, frames):
        assert atoms.get_potential_energy() == ref.get_potential_energy()
        assert atoms.calc is not None
        assert atoms.calc.results.keys() == {"energy"}
        assert atoms.info == {}
        assert atoms.arrays.keys() == {"numbers", "positions"}
        npt.assert_array_equal(atoms.get_atomic_numbers(), ref.get_atomic_numbers())
        npt.assert_array_equal(atoms.get_positions(), ref.get_positions())


@pytest.mark.parametrize(
    "dataset",
    [
        "s22_info_arrays_calc",
    ],
)
def test_restricted_read_forces(tmp_path, dataset, request):
    frames = request.getfixturevalue(dataset)
    io = znh5md.IO(tmp_path / "test.h5")
    io.extend(frames)

    rio = znh5md.IO(tmp_path / "test.h5", include=["position", "force"])
    assert len(rio) == len(rio[:])
    assert len(rio) == len(frames)

    for atoms, ref in zip(rio, frames):
        npt.assert_array_equal(atoms.get_forces(), ref.get_forces())
        assert atoms.calc is not None
        assert atoms.calc.results.keys() == {"forces"}
        assert atoms.info == {}
        assert atoms.arrays.keys() == {"numbers", "positions"}
        npt.assert_array_equal(atoms.get_atomic_numbers(), ref.get_atomic_numbers())
        npt.assert_array_equal(atoms.get_positions(), ref.get_positions())


@pytest.mark.parametrize(
    "dataset",
    [
        "s22_info_arrays_calc",
    ],
)
def test_restricted_read_info(tmp_path, dataset, request):
    frames = request.getfixturevalue(dataset)
    io = znh5md.IO(tmp_path / "test.h5")
    io.extend(frames)

    rio = znh5md.IO(tmp_path / "test.h5", include=["position", "mlip_energy"])
    assert len(rio) == len(rio[:])
    assert len(rio) == len(frames)

    for atoms, ref in zip(rio, frames):
        assert atoms.calc is None
        assert atoms.info.keys() == {"mlip_energy"}
        assert atoms.info["mlip_energy"] == ref.info["mlip_energy"]
        assert atoms.arrays.keys() == {"numbers", "positions"}
        npt.assert_array_equal(atoms.get_atomic_numbers(), ref.get_atomic_numbers())
        npt.assert_array_equal(atoms.get_positions(), ref.get_positions())


@pytest.mark.parametrize(
    "dataset",
    [
        "s22_info_arrays_calc",
    ],
)
def test_restricted_read_arrays(tmp_path, dataset, request):
    frames = request.getfixturevalue(dataset)
    io = znh5md.IO(tmp_path / "test.h5")
    io.extend(frames)

    rio = znh5md.IO(tmp_path / "test.h5", include=["position", "mlip_forces"])
    assert len(rio) == len(rio[:])
    assert len(rio) == len(frames)

    for atoms, ref in zip(rio, frames):
        assert atoms.calc is None
        assert atoms.info == {}
        assert atoms.arrays.keys() == {"numbers", "positions", "mlip_forces"}
        npt.assert_array_equal(atoms.get_atomic_numbers(), ref.get_atomic_numbers())
        npt.assert_array_equal(atoms.get_positions(), ref.get_positions())
        npt.assert_array_equal(atoms.arrays["mlip_forces"], ref.arrays["mlip_forces"])


# New tests for restricted writing


@pytest.mark.parametrize(
    "dataset",
    [
        "s22_info_arrays_calc",
    ],
)
def test_restricted_write_position(tmp_path, dataset, request):
    frames = request.getfixturevalue(dataset)
    # Write only 'position'
    io = znh5md.IO(tmp_path / "test.h5", include=["position"])
    io.extend(frames)

    # Read everything back
    rio = znh5md.IO(tmp_path / "test.h5")
    assert len(rio) == len(frames)

    for atoms, ref in zip(rio, frames):
        assert atoms.calc is None
        assert atoms.info == {}
        assert atoms.arrays.keys() == {
            "numbers",
            "positions",
        }  # 'numbers' is always present
        npt.assert_array_equal(atoms.get_atomic_numbers(), ref.get_atomic_numbers())
        npt.assert_array_equal(atoms.get_positions(), ref.get_positions())


@pytest.mark.parametrize(
    "dataset",
    [
        "s22_mixed_pbc_cell",
    ],
)
def test_restricted_write_box(tmp_path, dataset, request):
    frames = request.getfixturevalue(dataset)
    # Write only 'position' and 'box'
    io = znh5md.IO(tmp_path / "test.h5", include=["position", "box"])
    io.extend(frames)

    # Read everything back
    rio = znh5md.IO(tmp_path / "test.h5")
    assert len(rio) == len(frames)

    for atoms, ref in zip(rio, frames):
        assert atoms.calc is None
        assert atoms.info == {}
        assert atoms.arrays.keys() == {
            "numbers",
            "positions",
        }  # 'numbers' is always present
        npt.assert_array_equal(atoms.get_atomic_numbers(), ref.get_atomic_numbers())
        npt.assert_array_equal(atoms.get_positions(), ref.get_positions())
        npt.assert_array_equal(atoms.get_cell(), ref.get_cell())
        npt.assert_array_equal(atoms.get_pbc(), ref.get_pbc())


@pytest.mark.parametrize(
    "dataset",
    [
        "s22_info_arrays_calc",
    ],
)
def test_restricted_write_potential_energy(tmp_path, dataset, request):
    frames = request.getfixturevalue(dataset)
    # Write only 'position' and 'potential_energy'
    io = znh5md.IO(
        tmp_path / "test.h5", include=["position", "energy"]
    )  # TODO: either use ASE keys of H5MD keys but don't mix them
    io.extend(frames)

    # Read everything back
    rio = znh5md.IO(tmp_path / "test.h5")
    assert len(rio) == len(frames)

    for atoms, ref in zip(rio, frames):
        assert atoms.calc is not None
        assert atoms.calc.results.keys() == {"energy"}
        assert np.isclose(atoms.get_potential_energy(), ref.get_potential_energy())
        assert atoms.info == {}
        assert atoms.arrays.keys() == {"numbers", "positions"}
        npt.assert_array_equal(atoms.get_atomic_numbers(), ref.get_atomic_numbers())
        npt.assert_array_equal(atoms.get_positions(), ref.get_positions())


@pytest.mark.parametrize(
    "dataset",
    [
        "s22_info_arrays_calc",
    ],
)
def test_restricted_write_forces(tmp_path, dataset, request):
    frames = request.getfixturevalue(dataset)
    # Write only 'position' and 'force'
    io = znh5md.IO(
        tmp_path / "test.h5", include=["position", "forces"]
    )  # TODO: either use ASE keys of H5MD keys but don't mix them
    io.extend(frames)

    # Read everything back
    rio = znh5md.IO(tmp_path / "test.h5")
    assert len(rio) == len(frames)

    for atoms, ref in zip(rio, frames):
        assert atoms.calc is not None
        assert atoms.calc.results.keys() == {"forces"}
        npt.assert_array_equal(atoms.get_forces(), ref.get_forces())
        assert atoms.info == {}
        assert atoms.arrays.keys() == {"numbers", "positions"}
        npt.assert_array_equal(atoms.get_atomic_numbers(), ref.get_atomic_numbers())
        npt.assert_array_equal(atoms.get_positions(), ref.get_positions())


@pytest.mark.parametrize(
    "dataset",
    [
        "s22_info_arrays_calc",
    ],
)
def test_restricted_write_info(tmp_path, dataset, request):
    frames = request.getfixturevalue(dataset)
    # Write only 'position' and 'mlip_energy' from info
    io = znh5md.IO(tmp_path / "test.h5", include=["position", "mlip_energy"])
    io.extend(frames)

    # Read everything back
    rio = znh5md.IO(tmp_path / "test.h5")
    assert len(rio) == len(frames)

    for atoms, ref in zip(rio, frames):
        assert atoms.calc is None
        assert atoms.info.keys() == {"mlip_energy"}
        assert atoms.info["mlip_energy"] == ref.info["mlip_energy"]
        assert atoms.arrays.keys() == {"numbers", "positions"}
        npt.assert_array_equal(atoms.get_atomic_numbers(), ref.get_atomic_numbers())
        npt.assert_array_equal(atoms.get_positions(), ref.get_positions())


@pytest.mark.parametrize(
    "dataset",
    [
        "s22_info_arrays_calc",
    ],
)
def test_restricted_write_arrays(tmp_path, dataset, request):
    frames = request.getfixturevalue(dataset)
    # Write only 'position' and 'mlip_forces' from arrays
    io = znh5md.IO(tmp_path / "test.h5", include=["position", "mlip_forces"])
    io.extend(frames)

    # Read everything back
    rio = znh5md.IO(tmp_path / "test.h5")
    assert len(rio) == len(frames)

    for atoms, ref in zip(rio, frames):
        assert atoms.calc is None
        assert atoms.info == {}
        assert atoms.arrays.keys() == {"numbers", "positions", "mlip_forces"}
        npt.assert_array_equal(atoms.get_atomic_numbers(), ref.get_atomic_numbers())
        npt.assert_array_equal(atoms.get_positions(), ref.get_positions())
        npt.assert_array_equal(atoms.arrays["mlip_forces"], ref.arrays["mlip_forces"])



import znh5md


def test_creator(tmp_path):
    io = znh5md.IO(tmp_path / "test.h5")
    # These are the defaults
    assert io.creator == "znh5md"
    assert io.creator_version == znh5md.__version__



import ase
from scm.plams import XYZTrajectoryFile

from .abc import IOBase


class PLAMSIO(IOBase):
    def setup(self):
        pass

    def read(self) -> list[ase.Atoms]:
        # Read the trajectory file using PLAMS
        if self.format == "xyz":
            traj = XYZTrajectoryFile(self.filename)
            mol = traj.get_plamsmol()
        else:
            raise ValueError(f"Unsupported format: {self.format}")
        frames = []
        for i in range(traj.get_length()):
            crd, _ = traj.read_frame(i, molecule=mol)
            atoms = ase.Atoms(
                positions=crd,
                pbc=True,
            )
            frames.append(atoms)
        return frames

    def write(self, atoms: list[ase.Atoms]) -> None:
        raise NotImplementedError



import ase.build
import ase.collections
import numpy as np
import pytest
from ase.calculators.singlepoint import SinglePointCalculator


@pytest.fixture
def s22() -> list[ase.Atoms]:
    return list(ase.collections.s22)


@pytest.fixture
def s22_energy() -> list[ase.Atoms]:
    images = []
    for atoms in ase.collections.s22:
        calc = SinglePointCalculator(atoms, energy=np.random.rand())
        atoms.calc = calc
        images.append(atoms)
    return images


@pytest.fixture
def s22_energy_forces() -> list[ase.Atoms]:
    images = []
    for atoms in ase.collections.s22:
        calc = SinglePointCalculator(
            atoms, energy=np.random.rand(), forces=np.random.rand(len(atoms), 3)
        )
        atoms.calc = calc
        images.append(atoms)
    return images


@pytest.fixture
def s22_all_properties() -> list[ase.Atoms]:
    images = []
    for atoms in ase.collections.s22:
        # shapes taken from https://gitlab.com/ase/ase/-/blob/master/ase/outputs.py
        energy = np.random.rand()
        energies = np.random.rand(len(atoms))
        free_energy = np.random.rand()

        forces = np.random.rand(len(atoms), 3)
        stress = np.random.rand(6)
        stresses = np.random.rand(len(atoms), 6)

        dipole = np.random.rand(3)
        magmom = np.random.rand()
        magmoms = np.random.rand(len(atoms))

        dielectric_tensor = np.random.rand(3, 3)
        born_effective_charges = np.random.rand(len(atoms), 3)
        polarization = np.random.rand(3)

        calc = SinglePointCalculator(
            atoms,
            energy=energy,
            energies=energies,
            free_energy=free_energy,
            forces=forces,
            stress=stress,
            stresses=stresses,
            dipole=dipole,
            magmom=magmom,
            magmoms=magmoms,
            dielectric_tensor=dielectric_tensor,
            born_effective_charges=born_effective_charges,
            polarization=polarization,
        )

        atoms.calc = calc
        images.append(atoms)
    return images


@pytest.fixture
def s22_info_arrays_calc() -> list[ase.Atoms]:
    images = []
    for atoms in ase.collections.s22:
        atoms: ase.Atoms
        atoms.info.update(
            {
                "mlip_energy": np.random.rand(),
                "mlip_energy_2": np.random.rand(),
                "mlip_stress": np.random.rand(6),
                "collection": "s22",
                "metadata": {"author": "Jane Doe", "date": "2021-09-01"},
                "lst": [1, 2, 3],
                # "lst_str": ["Jane Doe", "John Doe"],
                # "lst_array": [np.random.rand(3), np.random.rand(3)],
            }
        )
        atoms.new_array("mlip_forces", np.random.rand(len(atoms), 3))
        atoms.new_array("mlip_forces_2", np.random.rand(len(atoms), 3))
        # atoms.arrays["arr_lst_arr"] = [np.random.rand(3) for _ in range(len(atoms))]
        # atoms.arrays["arr_lst"] = [[1, 2, 3] for _ in range(len(atoms))]
        # atoms.new_array("arr_str", np.array(["abc" for _ in range(len(atoms))]))
        atoms.set_velocities(np.random.rand(len(atoms), 3))
        calc = SinglePointCalculator(
            atoms, energy=np.random.rand(), forces=np.random.rand(len(atoms), 3)
        )
        atoms.calc = calc
        images.append(atoms)
    return images


@pytest.fixture
def s22_mixed_pbc_cell() -> list[ase.Atoms]:
    images = []
    for atoms in ase.collections.s22:
        atoms.set_pbc(np.random.rand(3) > 0.5)
        atoms.set_cell(np.random.rand(3, 3))
        # add something to info
        atoms.info["mlip_energy"] = np.random.rand()
        # add something to arrays
        atoms.new_array("mlip_forces", np.random.rand(len(atoms), 3))
        images.append(atoms)
    return images


@pytest.fixture
def s22_illegal_calc_results() -> list[ase.Atoms]:
    images = []
    for atoms in ase.collections.s22:
        atoms.calc = SinglePointCalculator(atoms)
        atoms.calc.results["mlip_energy"] = np.random.rand()
        atoms.calc.results["dict"] = {"author": "Jane Doe", "date": "2021-09-01"}
        atoms.calc.results["float"] = 3.14
        atoms.calc.results["int"] = 42
        atoms.calc.results["list"] = [1, 2, 3]
        atoms.calc.results["str"] = '{"author": "Jane Doe", "date": "2021-09-01"}'
        atoms.calc.results["list_array"] = [np.random.rand(3), np.random.rand(3)]
        atoms.calc.results["list_str"] = ["Jane Doe", "John Doe"]
        # atoms.calc.results["list_dict"] = [
        #     {"author": "Jane Doe", "date": "2021-09-01"},
        #     {"author": "John Doe", "date": "2021-09-02"},
        # ]
        images.append(atoms)
    return images


@pytest.fixture
def water() -> list[ase.Atoms]:
    """Get a dataset without positions."""
    return [ase.Atoms("H2O")]


@pytest.fixture
def s22_no_ascii() -> list[ase.Atoms]:
    images = []
    for atoms in ase.collections.s22:
        atoms.info["config"] = "βγ"
        images.append(atoms)
    return images


@pytest.fixture
def frames_with_residuenames() -> list[ase.Atoms]:
    water = ase.build.molecule("H2O")
    # typical PDB array data
    water.arrays["residuenames"] = np.array(["H2O"] * len(water))
    water.arrays["atomtypes"] = np.array(["γO", "βH", "βH"])

    ethane = ase.build.molecule("C2H6")
    ethane.arrays["residuenames"] = np.array(["C2H6"] * len(ethane))
    ethane.arrays["atomtypes"] = np.array(["γC", "βH", "βH", "βH", "βH", "βH"])
    return [water, ethane]


@pytest.fixture
def s22_info_arrays_calc_missing_inbetween() -> list[ase.Atoms]:
    images = []
    for atoms in ase.collections.s22:
        atoms: ase.Atoms
        if np.random.random() > 0.5:
            atoms.info.update({"mlip_energy": np.random.rand()})
        if np.random.random() > 0.5:
            atoms.new_array("mlip_forces", np.random.rand(len(atoms), 3))
        if np.random.random() > 0.5:
            calc = SinglePointCalculator(atoms)
            set_calc = False
            if np.random.random() > 0.5:
                calc.results["energy"] = np.random.rand()
                set_calc = True
            if np.random.random() > 0.5:
                calc.results["forces"] = np.random.rand(len(atoms), 3)
                set_calc = True
            if set_calc:
                atoms.calc = calc
        images.append(atoms)
    return images


@pytest.fixture
def s22_nested_calc() -> list[ase.Atoms]:
    images = []
    for atoms in ase.collections.s22:
        atoms: ase.Atoms
        atoms.calc = SinglePointCalculator(atoms)
        atoms.calc.results["forces"] = np.random.rand(len(atoms), 3)
        atoms.calc.results["forces_contributions"] = [
            [
                np.random.rand(len(atoms), 3),
                np.random.rand(len(atoms), 3),
            ]
        ]
        images.append(atoms)
    return images


@pytest.fixture
def full_water(water) -> ase.Atoms:
    """Get a dataset with full water molecules."""
    # add a calculator and info and arrays
    water = water[0]
    water.calc = SinglePointCalculator(
        water, energy=1.0, forces=np.zeros((len(water), 3))
    )
    water.info["smiles"] = "O"
    water.arrays["mlip_forces"] = np.zeros((len(water), 3))

    return water



import numpy as np
import numpy.testing as npt
from ase.build import molecule

import znh5md


def test_high_dim_data_arrays(tmp_path):
    water = molecule("H2O")
    water.arrays["descriptor"] = np.random.rand(3, 10, 5, 4, 3)

    io = znh5md.IO(tmp_path / "test.h5")
    io.append(water)

    io2 = znh5md.IO(tmp_path / "test.h5")
    water2 = io2[0]

    npt.assert_array_equal(water.arrays["descriptor"], water2.arrays["descriptor"])


def test_high_dim_data_info(tmp_path):
    water = molecule("H2O")
    water.info["descriptor"] = np.random.rand(5, 10, 4, 4, 3)

    io = znh5md.IO(tmp_path / "test.h5")
    io.append(water)

    io2 = znh5md.IO(tmp_path / "test.h5")
    water2 = io2[0]

    npt.assert_array_equal(water.info["descriptor"], water2.info["descriptor"])



import ase.build
import numpy as np

import znh5md


def test_append_new_calc(tmp_path, s22, s22_energy):
    io = znh5md.IO(tmp_path / "test.h5")

    io.extend(s22)
    io.extend(s22_energy)
    io.extend(s22)
    io.extend(s22_energy)
    io.extend(s22)
    io.extend(s22_energy)

    assert len(s22) == 22
    assert len(s22_energy) == 22
    assert len(io) == len(s22) * 3 + len(s22_energy) * 3

    assert len(list(io)) == len(s22) * 3 + len(s22_energy) * 3

    for a, b in zip(io[:22], s22):
        assert a.calc is None
        assert b.calc is None

    for a, b in zip(io[22:44], s22_energy):
        assert a.calc is not None
        assert b.calc is not None
        assert a.calc.results == b.calc.results

    for a, b in zip(io[44:], s22):
        assert a.calc is None
        assert b.calc is None

    for a, b in zip(io[66:], s22_energy):
        assert a.calc is not None
        assert b.calc is not None
        assert a.calc.results == b.calc.results

    for a, b in zip(io[88:], s22):
        assert a.calc is None
        assert b.calc is None

    for a, b in zip(io[110:], s22_energy):
        assert a.calc is not None
        assert b.calc is not None
        assert a.calc.results == b.calc.results


def test_append_new_keys_info(tmp_path):
    io = znh5md.IO(tmp_path / "test.h5")
    water = ase.build.molecule("H2O")
    water.info["key1"] = np.random.random()

    io.append(ase.build.molecule("H2O"))
    io.append(water)
    io.append(ase.build.molecule("H2O"))
    io.append(water)

    assert len(list(io)) == 4

    assert len(io) == 4
    assert "key1" not in io[0].info
    assert "key1" in io[1].info
    assert "key1" not in io[2].info
    assert "key1" in io[3].info

    assert io[1].info["key1"] == water.info["key1"]


def test_append_new_keys_info_non_ascii(tmp_path):
    io = znh5md.IO(tmp_path / "test.h5")
    water = ase.build.molecule("H2O")
    water.info["key1"] = "γC"

    io.append(ase.build.molecule("H2O"))
    io.append(water)
    io.append(ase.build.molecule("H2O"))
    io.append(water)

    assert len(list(io)) == 4

    assert len(io) == 4
    assert "key1" not in io[0].info
    assert "key1" in io[1].info
    assert "key1" not in io[2].info
    assert "key1" in io[3].info

    assert io[1].info["key1"] == "γC" == io[3].info["key1"]


def test_append_new_keys_arrays(tmp_path):
    io = znh5md.IO(tmp_path / "test.h5")
    water = ase.build.molecule("H2O")
    water.arrays["key1"] = np.random.rand(len(water), 3)

    io.append(ase.build.molecule("H2O"))
    io.append(water)
    io.append(ase.build.molecule("H2O"))
    io.append(water)

    assert len(list(io)) == 4

    assert len(io) == 4
    assert "key1" not in io[0].arrays
    assert "key1" in io[1].arrays
    assert "key1" not in io[2].arrays
    assert "key1" in io[3].arrays

    assert np.array_equal(io[1].arrays["key1"], water.arrays["key1"])



import numpy as np
import numpy.testing as npt
import pytest

from znh5md import serialization as zns


def test_serialisation_s22(s22):
    frames = zns.Frames.from_ase(s22)

    assert len(frames.positions) == 22
    assert len(frames.numbers) == 22
    assert len(frames.pbc) == 22
    assert len(frames.cell) == 22
    assert len(frames.arrays) == 0
    assert len(frames.info) == 0
    assert len(frames.calc) == 0


@pytest.mark.parametrize(
    "dataset_name",
    [
        "s22",
        "s22_energy",
        "s22_energy_forces",
        "s22_all_properties",
        "s22_info_arrays_calc",
        "s22_mixed_pbc_cell",
        "s22_illegal_calc_results",
        "s22_no_ascii",
        "frames_with_residuenames",
    ],
)
@pytest.mark.parametrize("append", [True, False])
def test_frames_iter(dataset_name, append, request):
    dataset = request.getfixturevalue(dataset_name)
    frames = zns.Frames()
    if append:
        for frame in dataset:
            frames.append(frame)
    else:
        frames.extend(dataset)

    for a, b in zip(frames, dataset):
        npt.assert_array_equal(a.positions, b.positions)
        npt.assert_array_equal(a.cell, b.cell)
        npt.assert_array_equal(a.pbc, b.pbc)
        npt.assert_array_equal(a.numbers, b.numbers)

        for key in b.arrays:
            npt.assert_array_equal(a.arrays[key], b.arrays[key])
        for key in b.info:
            npt.assert_array_equal(a.info[key], b.info[key])
        if b.calc is not None:
            for key in b.calc.results:
                npt.assert_array_equal(a.calc.results[key], b.calc.results[key])


@pytest.mark.parametrize(
    "dataset_name",
    [
        "s22",
        "s22_energy",
        "s22_energy_forces",
        "s22_all_properties",
        "s22_info_arrays_calc",
        "s22_mixed_pbc_cell",
        "s22_illegal_calc_results",
        "s22_no_ascii",
        "frames_with_residuenames",
        "s22_info_arrays_calc_missing_inbetween",
    ],
)
def test_frames_getitem(dataset_name, request):
    dataset = request.getfixturevalue(dataset_name)
    frames = zns.Frames.from_ase(dataset)

    for idx, frame in enumerate(dataset):
        a = frames[idx]
        b = frame

        npt.assert_array_equal(a.positions, b.positions)
        npt.assert_array_equal(a.cell, b.cell)
        npt.assert_array_equal(a.pbc, b.pbc)
        npt.assert_array_equal(a.numbers, b.numbers)

        for key in b.arrays:
            npt.assert_array_equal(a.arrays[key], b.arrays[key])
        for key in b.info:
            npt.assert_array_equal(a.info[key], b.info[key])
        if b.calc is not None:
            for key in b.calc.results:
                npt.assert_array_equal(a.calc.results[key], b.calc.results[key])


def test_len(s22):
    frames = zns.Frames.from_ase(s22)

    assert len(frames) == 22

    assert len(list(frames)) == 22


def test_serialisation_s22_energy(s22_energy):
    frames = zns.Frames.from_ase(s22_energy)

    assert len(frames.arrays) == 0
    assert len(frames.info) == 0
    assert len(frames.calc) == 1
    assert len(frames.calc["energy"]) == 22
    assert all(isinstance(x, float) for x in frames.calc["energy"])


def test_serialisation_s22_energy_forces(s22_energy_forces):
    frames = zns.Frames.from_ase(s22_energy_forces)

    assert len(frames.arrays) == 0
    assert len(frames.info) == 0
    assert len(frames.calc) == 2
    assert len(frames.calc["energy"]) == 22
    assert all(isinstance(x, float) for x in frames.calc["energy"])
    assert len(frames.calc["forces"]) == 22
    assert all(isinstance(x, np.ndarray) for x in frames.calc["forces"])



import ase
import MDAnalysis as mda  # noqa: N813
from MDAnalysis.coordinates.H5MD import H5MDReader
from MDAnalysis.coordinates.XTC import XTCReader

from .abc import IOBase


class MDAIO(IOBase):
    def setup(self):
        pass

    def read(self) -> list[ase.Atoms]:
        if self.format == "h5md":
            universe = mda.Universe.empty(self.num_atoms, trajectory=True)
            reader = H5MDReader(
                self.filename,
                convert_units=False,
                # dt=2,
                # time_offset=10,
                # foo="bar",
            )
            universe.trajectory = reader
        elif self.format == "xtc":
            universe = mda.Universe.empty(self.num_atoms, trajectory=True)
            reader = XTCReader(
                self.filename,
                convert_units=False,
            )
            universe.trajectory = reader
        else:
            universe = mda.Universe(self.filename, format=self.format.upper())

        universe.transfer_to_memory()
        frames = []
        for ts in universe.trajectory:
            positions = ts.positions
            cell_vectors = ts.dimensions
            atoms = ase.Atoms(positions=positions, cell=cell_vectors, pbc=True)
            frames.append(atoms)
        return frames

    def write(self, atoms: list[ase.Atoms]) -> None:
        raise NotImplementedError



import numpy as np
import numpy.testing as npt
import pytest

from znh5md.misc import concatenate_varying_shape_arrays, decompose_varying_shape_arrays


@pytest.mark.parametrize(
    ("inp", "expected", "fillvalue"),
    [
        # 0D examples
        ([np.array(1), np.array(2)], np.array([1, 2]), 0),
        # 1D examples
        (
            [np.array([1, 2]), np.array([3, 4, 5])],
            np.array([[1, 2, np.nan], [3, 4, 5]]),
            np.nan,
        ),
        ([np.array([1, 2]), np.array([3, 4])], np.array([[1, 2], [3, 4]]), 0),
        (
            [np.array([1, 2, 3]), np.array([4, 5])],
            np.array([[1, 2, 3], [4, 5, np.nan]]),
            np.nan,
        ),
        # Higher-dimensional examples
        (
            [np.array([[1, 2], [3, 4]]), np.array([[5, 6]])],
            np.array([[[1, 2], [3, 4]], [[5, 6], [np.nan, np.nan]]]),
            np.nan,
        ),
        (
            [np.array([[1, 2], [3, 4]]), np.array([[5, 6], [7, 8]])],
            np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]]),
            0,
        ),
    ],
)
def test_concatenate_varying_shape_arrays(inp, expected, fillvalue):
    result = concatenate_varying_shape_arrays(
        inp, fillvalue=fillvalue, dtype=np.float64
    )
    npt.assert_array_equal(result, expected)


@pytest.mark.parametrize(
    ("inp", "shape"),
    [
        [[np.random.rand(3), np.random.rand(2)], (2, 3)],
        [[np.random.rand(10, 3), np.random.rand(20, 3)], (2, 20, 3)],
        [[np.random.rand(10, 10, 3), np.random.rand(20, 10, 3)], (2, 20, 10, 3)],
        [[np.random.rand(10, 10, 3), np.random.rand(10, 20, 3)], (2, 10, 20, 3)],
        [[np.random.rand(20, 10, 3), np.random.rand(10, 20, 3)], (2, 20, 20, 3)],
        [
            [
                np.random.rand(20, 10, 3),
                np.random.rand(10, 20, 3),
                np.random.rand(10, 20, 4),
            ],
            (3, 20, 20, 4),
        ],
    ],
)
def test_concatenate_varying_shape_arrays_concept(inp, shape):
    result = concatenate_varying_shape_arrays(inp, fillvalue=np.nan, dtype=np.float64)
    assert result.shape == shape


@pytest.mark.parametrize(
    "inp",
    [
        [np.random.rand(3), np.random.rand(2)],
        [np.random.rand(10, 3), np.random.rand(20, 3)],
        [np.random.rand(10, 10, 3), np.random.rand(20, 10, 3)],
        [np.random.rand(10, 10, 3), np.random.rand(10, 20, 3)],
        [np.random.rand(20, 10, 3), np.random.rand(10, 20, 3)],
        [
            np.random.rand(20, 10, 3),
            np.random.rand(10, 20, 3),
            np.random.rand(10, 20, 4),
        ],
    ],
)
@pytest.mark.parametrize("fillvalue", [np.nan, 0.0])
def test_decompose_varying_shape_arrays(inp, fillvalue):
    concatenated = concatenate_varying_shape_arrays(
        inp, fillvalue=fillvalue, dtype=np.float64
    )
    result = decompose_varying_shape_arrays(concatenated, fillvalue=fillvalue)
    assert len(result) == len(inp)
    for i, array in enumerate(result):
        assert array.shape == inp[i].shape
        npt.assert_array_equal(array, inp[i])



from dataclasses import dataclass

import ase


@dataclass(kw_only=True)
class IOBase:
    filename: str
    format: str
    num_atoms: int
    num_frames: int

    def setup(self):
        pass

    def read(self) -> list[ase.Atoms]: ...

    def write(self, atoms: list[ase.Atoms]) -> None: ...



from ase.atoms import Atoms
from ase.collections import s22
from ase.io import read, write

import znh5md


def test_write_one(tmp_path):
    path = tmp_path / "test.h5"

    atoms = list(s22)[0]
    write(path, atoms)

    io = znh5md.IO(path)
    images_read = io[:]

    assert len(images_read) == 1
    assert atoms == images_read[0]


def test_write_multiple(tmp_path):
    path = tmp_path / "test.h5"

    images = list(s22)
    write(path, images)

    io = znh5md.IO(path)
    images_read = io[:]

    assert len(images) == len(images_read)
    for image, image_read in zip(images, images_read):
        assert image == image_read


def test_append(tmp_path):
    path = tmp_path / "test.h5"

    images = list(s22)
    for image in images:
        write(path, image, append=True)

    io = znh5md.IO(path)
    images_read = io[:]

    assert len(images) == len(images_read)
    for image, image_read in zip(images, images_read):
        assert image == image_read


def test_read_one(tmp_path):
    path = tmp_path / "test.h5"
    io = znh5md.IO(path)

    images = list(s22)
    io.extend(images)

    image_read = read(path, index=0)
    assert isinstance(image_read, Atoms)
    assert images[0] == image_read


def test_read_multiple(tmp_path):
    path = tmp_path / "test.h5"
    io = znh5md.IO(path)

    images = list(s22)
    io.extend(images)

    images_read = read(path, index=":")
    assert isinstance(images_read, list)
    for image, image_read in zip(images, images_read):
        assert image == image_read



import os

import mdtraj as md
import numpy as np
import pandas as pd
from ase import Atoms
from ase.io.proteindatabank import write_proteindatabank

from .abc import IOBase


class MDTrajIO(IOBase):
    def setup(self):
        atom_names = [f"H{i + 1}" for i in range(self.num_atoms)]  # Unique atom names
        elements = [
            "H"
        ] * self.num_atoms  # All elements are H as per original user code structure

        top_df = pd.DataFrame(
            {
                "name": atom_names,  # Atom names (e.g., H1, H2)
                "element": elements,  # Element symbols (e.g., H, C, O)
                "resSeq": [1] * self.num_atoms,
                "resName": ["MOL"] * self.num_atoms,  # Using a generic residue name
                "chainID": ["A"] * self.num_atoms,
                "serial": np.arange(self.num_atoms),
            }
        )
        self.topology = md.Topology.from_dataframe(top_df)

    def read(self) -> list[Atoms]:
        # Read the trajectory file using MDTraj
        if self.format == "xyz":
            traj = md.load_xyz(self.filename, top=self.topology)
            frames = []
            for i in range(traj.n_frames):
                positions = traj.xyz[i]
                atoms = Atoms(
                    positions=positions,
                    pbc=True,
                )
                frames.append(atoms)
            return frames
        elif self.format == "pdb":
            traj = md.load_pdb(self.filename)
            frames = []
            for i in range(traj.n_frames):
                positions = traj.xyz[i]
                atoms = Atoms(
                    positions=positions,
                    pbc=True,
                )
                frames.append(atoms)
            return frames
        elif self.format == "xtc":
            traj = md.load_xtc(self.filename, top=self.topology)
            frames = []
            for i in range(traj.n_frames):
                positions = traj.xyz[i]
                atoms = Atoms(
                    positions=positions,
                    pbc=True,
                )
                frames.append(atoms)
            return frames
        else:
            raise ValueError(f"Unsupported format: {self.format}")

    def write(self, frames: list[Atoms]):
        if self.format == "xtc":
            # use ase to write xyz, load that file and then save as xtc
            write_proteindatabank(
                "ase_pdb_to_xtc_tmp.pdb", frames
            )  # Write to a temporary XYZ file
            traj = md.load_pdb("ase_pdb_to_xtc_tmp.pdb")
            traj.save_xtc(self.filename)
            # remove the temporary file
            os.remove("ase_pdb_to_xtc_tmp.pdb")
        else:
            raise ValueError(f"Unsupported format: {self.format}")



"""Test against MDAnalysis."""

import ase.build
import MDAnalysis as mda  # noqa:N813
import numpy as np
import numpy.testing as npt
from ase.calculators.singlepoint import SinglePointCalculator
from MDAnalysis.coordinates.H5MD import H5MDReader

import znh5md


def test_mda_read(tmp_path):
    water = ase.build.molecule("H2O")
    water.set_momenta(np.random.rand(*water.positions.shape))
    water.calc = SinglePointCalculator(
        water, forces=np.random.rand(*water.positions.shape)
    )
    io = znh5md.IO(tmp_path / "test.h5", save_units=False, store="time")
    io.append(water)
    io.append(water)

    u = mda.Universe.empty(3, trajectory=True)
    reader = H5MDReader(tmp_path / "test.h5", convert_units=False)
    u.trajectory = reader
    for idx, ts in enumerate(u.trajectory):
        assert ts.frame == idx
        npt.assert_allclose(ts.positions, water.positions)
        npt.assert_allclose(ts.velocities, water.get_velocities())
        npt.assert_allclose(ts.forces, water.calc.results["forces"])


def test_mda_read_convert(tmp_path):
    water = ase.build.molecule("H2O")
    io = znh5md.IO(tmp_path / "test.h5", store="time")
    io.append(water)
    io.append(water)
    u = mda.Universe.empty(3, trajectory=True)
    reader = H5MDReader(tmp_path / "test.h5", convert_units=True)
    u.trajectory = reader
    for idx, ts in enumerate(u.trajectory):
        assert ts.frame == idx
        npt.assert_allclose(ts.positions, water.positions)



import contextlib
import dataclasses
import functools
import json
import typing as t

import ase
import h5py
import numpy as np
from ase.calculators.singlepoint import SinglePointCalculator

from znh5md.misc import MISSING, concatenate_varying_shape_arrays
from znh5md.units import get_unit

# Define allowed types
ALLOWED_TYPES = t.Union[np.ndarray, dict, float, int, str, bool, list, type(MISSING)]

# Define content type
CONTENT_TYPE = dict[str, ALLOWED_TYPES]

# Define origin type
ORIGIN_TYPE = t.Literal["calc", "info", "arrays", "atoms"]


@dataclasses.dataclass
class Entry:
    value: list[ALLOWED_TYPES] | np.ndarray = dataclasses.field(repr=False)
    origin: ORIGIN_TYPE | None
    name: str
    unit: str | None = None

    def __post_init__(self):
        if self.unit is None:
            self.unit = get_unit(self.name)

        if isinstance(self.value, list):
            if isinstance(self.value[0], list):
                try:
                    self.value = [np.array(v, dtype=np.float64) for v in self.value]
                except ValueError:
                    pass

    @functools.cached_property
    def ref(self) -> t.Any:
        for v in self.value:
            if v is not MISSING:
                return v
        else:
            raise ValueError("All values are MISSING")

    @functools.cached_property
    def dtype(self) -> t.Any:
        if isinstance(self.ref, np.ndarray):
            if self.ref.dtype.kind in ["O", "S", "U"]:
                return h5py.string_dtype()
            else:
                return np.float64
        elif isinstance(self.ref, str):
            return h5py.string_dtype()
        elif isinstance(self.ref, dict):
            return h5py.string_dtype()
        elif isinstance(self.ref, list):
            if any(not isinstance(v, (int, float, bool)) for v in self.ref):
                return h5py.string_dtype()
            return np.float64
        else:
            return np.float64

    @property
    def fillvalue(self) -> t.Any:
        if self.dtype == h5py.string_dtype():
            return ""
        else:
            return np.nan

    def dump(self) -> t.Tuple[np.ndarray | list, t.Any]:
        data = self.value
        try:
            if self.dtype == h5py.string_dtype():
                # Handle string data
                serialized_data = [
                    json.dumps(v.tolist() if isinstance(v, np.ndarray) else v)
                    if v is not MISSING
                    else ""
                    for v in data
                ]
                return serialized_data, h5py.string_dtype()
            else:
                # Handle non-string data
                processed_data = [
                    np.array(v, dtype=self.dtype)
                    if v is not MISSING
                    else np.full_like(self.ref, self.fillvalue, dtype=self.dtype)
                    for v in data
                ]
                return (
                    concatenate_varying_shape_arrays(
                        processed_data, self.fillvalue, dtype=np.float64
                    ),
                    np.float64,
                )
        except:
            print(f"Error in dump for {self}")
            print(self.value)
            print(type(self.value))
            raise


def process_momenta(target: dict[str, list], atoms: ase.Atoms, index: int) -> None:
    if "velocities" not in target:
        if "momenta" in atoms.arrays:
            target["velocities"] = [MISSING] * index + [atoms.get_velocities()]
    else:
        if "momenta" in atoms.arrays:
            target["velocities"].append(atoms.get_velocities())
        else:
            target["velocities"].append(MISSING)


def process_category(
    target: dict[str, list], content: CONTENT_TYPE, index: int
) -> None:
    """
    Process a category (arrays, info, calc) for a single atoms object, ensuring
    that keys are added and missing values are backfilled.

    Parameters
    ----------
    target : dict
        The target dictionary to update.
    content : dict
        The content to add to the target dictionary.
    index : int
        The index of the current frame.
    """
    seen = set(content.keys()) | {"momenta", "velocities"}
    unseen = set(target.keys()) - seen

    for key in content:
        if key == "momenta":
            # Momenta are handled separately via velocities
            continue
        if key not in target:
            # Backfill existing entries with MISSING for the new key
            target[key] = [MISSING] * index + [content[key]]
        else:
            # Add the new data to the existing key
            target[key].append(content[key])

    for key in unseen:
        # Backfill missing entries with MISSING for the unseen key
        target[key].append(MISSING)


# TODO: provide a reference for each, for the later padding to write to h5


@dataclasses.dataclass(repr=False)
class Frames:
    """Dataclass for Atoms object serialization."""

    positions: list = dataclasses.field(default_factory=list)
    numbers: list = dataclasses.field(default_factory=list)
    pbc: list = dataclasses.field(default_factory=list)
    cell: list = dataclasses.field(default_factory=list)
    arrays: dict[str, list] = dataclasses.field(default_factory=dict)
    info: dict[str, list] = dataclasses.field(default_factory=dict)
    calc: dict[str, list] = dataclasses.field(default_factory=dict)

    @classmethod
    def from_ase(
        cls, frames: t.Iterable[ase.Atoms], include: list[str] | None = None
    ) -> "Frames":
        """Create a Frames object from a sequence of ASE Atoms objects."""
        obj = cls()
        obj.extend(frames, include=include)
        return obj

    def keys(self) -> t.Iterator[str]:
        """Iterate over the keys."""
        yield "numbers"
        yield "positions"
        yield "pbc"
        yield "cell"
        for key in self.arrays:
            yield key
        for key in self.info:
            yield key
        for key in self.calc:
            yield key

    def items(self) -> t.Iterator[Entry]:
        """Iterate over the items."""
        yield Entry(self.numbers, "atoms", name="numbers")  # numbers has to be first!
        if len(self.positions) > 0:
            yield Entry(self.positions, "atoms", name="positions")
        if len(self.pbc) > 0:
            yield Entry(self.pbc, "atoms", name="pbc")
        if len(self.cell) > 0:
            yield Entry(self.cell, "atoms", name="cell")
        for key in self.arrays:
            yield Entry(self.arrays[key], "arrays", name=key)
        for key in self.info:
            yield Entry(self.info[key], "info", name=key)
        for key in self.calc:
            yield Entry(self.calc[key], "calc", name=key)

    def check(self) -> None:
        if len(list(self.keys())) != len(set(self.keys())):
            # find duplicates
            duplicates = []
            seen = set()
            for key in self.keys():
                if key in seen:
                    duplicates.append(key)
                else:
                    seen.add(key)
            raise ValueError(f"Duplicate keys found in Frames object: {duplicates}")

    def __iter__(self) -> t.Iterator[ase.Atoms]:
        """Iterate over the frames."""
        if self.numbers is None:
            return iter([])
        for idx in range(len(self.numbers)):
            try:
                yield self[idx]
            except IndexError:
                return

    def __len__(self) -> int:
        """Return the number of frames."""
        if self.numbers is None:
            return 0
        return len(self.numbers)

    def __getitem__(self, idx: int) -> ase.Atoms:  # noqa: C901
        """Return a single frame."""
        # this raises the IndexError to determine the length of the Frames object

        if self.numbers[idx] is MISSING:
            atoms = ase.Atoms()
        else:
            atoms = ase.Atoms(
                numbers=self.numbers[idx],
            )
        if len(self.positions) > 0:
            if self.positions[idx] is not MISSING:
                atoms.set_positions(self.positions[idx])
        if len(self.cell) > 0:
            atoms.set_cell(self.cell[idx])
        if len(self.pbc) > 0:
            atoms.set_pbc(self.pbc[idx])
        # all data following here can be missing
        for key in self.arrays:
            with contextlib.suppress(IndexError):
                if self.arrays[key][idx] is MISSING:
                    continue
                if key == "velocities":
                    atoms.set_velocities(self.arrays[key][idx])
                else:
                    atoms.arrays[key] = self.arrays[key][idx]

        for key in self.info:
            with contextlib.suppress(IndexError):
                if self.info[key][idx] is not MISSING:
                    atoms.info[key] = self.info[key][idx]
        for key in self.calc:
            with contextlib.suppress(IndexError):
                if self.calc[key][idx] is not MISSING:
                    if atoms.calc is None:
                        atoms.calc = SinglePointCalculator(atoms)
                    atoms.calc.results[key] = self.calc[key][idx]

        return atoms

    def append(self, atoms: ase.Atoms) -> None:
        """Append a frame to the frames."""
        self.extend([atoms])

    def extend(  # noqa: C901
        self, frames: t.Iterable[ase.Atoms], include: list[str] | None = None
    ) -> None:
        """Extend the frames with a sequence of frames."""

        if include is not None:
            # Convert include list to a set for faster lookups
            include_set = set(include)
        else:
            include_set = None

        start_idx = len(self.positions)
        self.numbers.extend([atoms.numbers for atoms in frames])

        if include_set is None or "position" in include_set:
            self.positions.extend([atoms.positions for atoms in frames])
        if include_set is None or "box" in include_set:
            self.pbc.extend([atoms.pbc for atoms in frames])
            self.cell.extend([atoms.cell.array for atoms in frames])

        for idx, atoms in enumerate(frames, start=start_idx):
            # Process arrays
            if include_set is None:
                atoms_arrays = {
                    key: value
                    for key, value in atoms.arrays.items()
                    if key not in ["positions", "numbers"]
                }
            else:
                atoms_arrays = {
                    key: value
                    for key, value in atoms.arrays.items()
                    if key not in ["positions", "numbers"] and key in include
                }
            process_category(self.arrays, atoms_arrays, idx)
            if include_set is None or "velocities" in include_set:
                process_momenta(self.arrays, atoms, idx)

            # Process info
            if include_set is None:
                atoms_info = atoms.info
            else:
                atoms_info = {
                    key: value
                    for key, value in atoms.info.items()
                    if key in include_set
                }
            process_category(self.info, atoms_info, idx)

            # Process calc
            if atoms.calc is not None:
                if include_set is None:
                    atoms_calc = atoms.calc.results
                else:
                    atoms_calc = {
                        key: value
                        for key, value in atoms.calc.results.items()
                        if key in include_set
                    }
                process_category(self.calc, atoms_calc, idx)
            elif len(self.calc) != 0:
                process_category(self.calc, {}, idx)



import ase
import numpy as np


def create_frames(num_frames=10, num_atoms=5) -> list[ase.Atoms]:
    """
    Create a random trajectory with n_frames and n_atoms using ASE.
    Each frame will have a random selection of H, O, C atoms.
    """
    frames = []
    for _ in range(num_frames):
        symbols = np.random.choice(["H", "O", "C"], size=num_atoms)
        positions = np.random.rand(num_atoms, 3) * 10
        cell_vectors = np.eye(3) * 100
        atoms = ase.Atoms(
            symbols=symbols, positions=positions, cell=cell_vectors, pbc=False
        )
        frames.append(atoms)
    return frames



import ase.build
import h5py
import numpy.testing as npt
import pytest

import znh5md


@pytest.mark.parametrize("pbc_group", [True, False])
def test_pbc_group_water(tmp_path, pbc_group):
    io = znh5md.IO(tmp_path / "test.h5", pbc_group=pbc_group)
    io.append(ase.build.molecule("H2O"))

    io2 = znh5md.IO(tmp_path / "test.h5")
    assert io2[0] == io[0]
    assert io2[0] == ase.build.molecule("H2O")
    npt.assert_array_equal(io2[0].get_pbc(), io[0].get_pbc())

    with h5py.File(tmp_path / "test.h5", "r") as f:
        if pbc_group:
            assert "pbc" in f["particles/atoms/box"]
        else:
            assert "pbc" not in f["particles/atoms/box"]



import ase
import numpy as np
import numpy.testing as npt
import pytest

import znh5md


def test_write_single(tmp_path):
    atoms = ase.Atoms("H2O", positions=np.random.rand(3, 3))
    file = tmp_path / "test.h5"
    znh5md.write(file, atoms)
    assert file.exists()


def test_read_write(tmp_path, s22_all_properties):
    file = tmp_path / "test.h5"
    znh5md.write(file, s22_all_properties)
    atoms = znh5md.read(file)
    assert atoms == s22_all_properties[-1]
    npt.assert_array_equal(
        atoms.get_positions(), s22_all_properties[-1].get_positions()
    )

    assert len(znh5md.read(file, index=slice(None, None, None))) == len(
        s22_all_properties
    )
    assert len(znh5md.read(file, index=[0, 1, 2])) == 3


def test_iread(tmp_path, s22_all_properties):
    file = tmp_path / "test.h5"
    znh5md.write(file, s22_all_properties)

    assert len(list(znh5md.iread(file))) == len(s22_all_properties)
    assert len(list(znh5md.IO(file))) == len(s22_all_properties)

    for a, b in zip(s22_all_properties, znh5md.iread(file)):
        npt.assert_array_equal(a.get_positions(), b.get_positions())
        npt.assert_array_equal(a.get_atomic_numbers(), b.get_atomic_numbers())


def test_append_false(tmp_path, s22_all_properties):
    file = tmp_path / "test.h5"
    znh5md.write(file, s22_all_properties)
    with pytest.raises(FileExistsError):
        znh5md.write(file, s22_all_properties, append=False)


def test_write_H(tmp_path):
    n_atoms = 10
    images = [
        ase.Atoms("H" * n_atoms, positions=np.random.rand(n_atoms, 3))
        for _ in range(10000)
    ]
    file = tmp_path / "test.h5"
    znh5md.write(file, images)
    atoms = znh5md.read(file)
    assert atoms == images[-1]



from dataclasses import dataclass

import ase

import znh5md

from .abc import IOBase


@dataclass(kw_only=True)
class ZnH5MDIO(IOBase):
    compression: str | None = "gzip"

    def setup(self):
        pass

    def read(self) -> list[ase.Atoms]:
        if self.format == "h5md":
            return znh5md.IO(self.filename)[:]
        else:
            raise ValueError(f"Unsupported format: {self.format}")

    def write(self, atoms: list[ase.Atoms]) -> None:
        if self.format == "h5md":
            znh5md.IO(self.filename, store="time", compression=self.compression).extend(
                atoms
            )
        else:
            raise ValueError(f"Unsupported format: {self.format}")


class ZnH5MDFixedShapeIO(IOBase):
    def setup(self):
        pass

    def read(self) -> list[ase.Atoms]:
        if self.format == "h5md":
            return znh5md.IO(self.filename, variable_shape=False)[:]
        else:
            raise ValueError(f"Unsupported format: {self.format}")

    def write(self, atoms: list[ase.Atoms]) -> None:
        if self.format == "h5md":
            znh5md.IO(self.filename, store="time", variable_shape=False).extend(atoms)
        else:
            raise ValueError(f"Unsupported format: {self.format}")



import numpy as np
import numpy.testing as npt
import pytest
from ase.build import molecule
from ase.calculators.singlepoint import SinglePointCalculator

import znh5md


# Define assertion functions for different data types
def assert_equal(actual, expected):
    npt.assert_equal(actual, expected)


def assert_allclose(actual, expected):
    npt.assert_allclose(actual, expected)


def assert_allclose_list(actual, expected):
    for a, e in zip(actual, expected):
        npt.assert_allclose(a, e)


def assert_dict_allclose(actual, expected):
    for key in expected:
        npt.assert_allclose(actual[key], expected[key])


# Pytest fixtures
@pytest.fixture
def io_fixture(tmp_path):
    """Fixture to initialize the ZnH5MD IO object."""
    return znh5md.IO(tmp_path / "test.h5")


@pytest.fixture
def water_molecule():
    """Fixture to provide a simple water molecule."""
    return molecule("H2O")


@pytest.mark.parametrize(
    "key,value,assert_fn",
    [
        ("str", "Hello World", assert_equal),
        ("float", 3.14, assert_equal),
        ("ndarray", np.random.rand(10), assert_allclose),
        ("list_float", [1.0, 2.0, 3.0], assert_allclose),
        ("list_str", ["Hello", "World"], assert_equal),
        ("dict", {"a": 1, "b": 2}, assert_equal),
        ("dict_str", {"a": "Hello", "b": "World"}, assert_equal),
        ("list_array", [np.random.rand(10), np.random.rand(10)], assert_allclose_list),
        ("dict_dict", {"a": {"x": 1, "y": 2}, "b": {"x": 3, "y": 4}}, assert_equal),
        # (
        #     "dict_array",
        #     {"a": np.random.rand(10), "b": np.random.rand(10)},
        #     assert_dict_allclose,
        # ),
        # ("list_dict", [{"a": 1, "b": 2}, {"c": 3, "d": 4}], assert_equal),
    ],
)
def test_info(io_fixture, water_molecule, key, value, assert_fn):
    """Generic test for different info types."""
    # Assign the value to the molecule's info
    water_molecule.info[key] = value

    # Append to the ZnH5MD IO object
    io_fixture.append(water_molecule)

    # Retrieve and test
    assert_fn(io_fixture[0].info[key], value)
    assert key not in io_fixture[0].arrays
    assert io_fixture[0].calc is None


@pytest.mark.parametrize(
    "key,value,assert_fn",
    [
        ("str", "Hello World", assert_equal),
        ("float", 3.14, assert_equal),
        ("ndarray", np.random.rand(10), assert_allclose),
        ("list_float", [1.0, 2.0, 3.0], assert_allclose),
        ("list_str", ["Hello", "World"], assert_equal),
        ("dict", {"a": 1, "b": 2}, assert_equal),
        ("dict_str", {"a": "Hello", "b": "World"}, assert_equal),
        ("list_array", [np.random.rand(10), np.random.rand(10)], assert_allclose_list),
        ("dict_dict", {"a": {"x": 1, "y": 2}, "b": {"x": 3, "y": 4}}, assert_equal),
        # (
        #     "dict_array",
        #     {"a": np.random.rand(10), "b": np.random.rand(10)},
        #     assert_dict_allclose,
        # ),
        # ("list_dict", [{"a": 1, "b": 2}, {"c": 3, "d": 4}], assert_equal),
    ],
)
def test_calc(io_fixture, water_molecule, key, value, assert_fn):
    """Generic test for different calc types."""
    # Assign the value to the molecule's info
    water_molecule.calc = SinglePointCalculator(water_molecule)
    water_molecule.calc.results[key] = value

    # Append to the ZnH5MD IO object
    io_fixture.append(water_molecule)

    # Retrieve and test
    assert_fn(io_fixture[0].calc.results[key], value)
    assert key not in io_fixture[0].arrays
    assert io_fixture[0].info == {}


@pytest.mark.parametrize(
    "key,value,assert_fn",
    [
        ("str", "Hello World", assert_equal),
        ("float", 3.14, assert_equal),
        ("ndarray", np.random.rand(10), assert_allclose),
        ("list_float", [1.0, 2.0, 3.0], assert_allclose),
        ("list_str", ["Hello", "World"], assert_equal),
        ("dict", {"a": 1, "b": 2}, assert_equal),
        ("dict_str", {"a": "Hello", "b": "World"}, assert_equal),
        ("list_array", [np.random.rand(10), np.random.rand(10)], assert_allclose_list),
        ("dict_dict", {"a": {"x": 1, "y": 2}, "b": {"x": 3, "y": 4}}, assert_equal),
        # (
        #     "dict_array",
        #     {"a": np.random.rand(10), "b": np.random.rand(10)},
        #     assert_dict_allclose,
        # ),
        # ("list_dict", [{"a": 1, "b": 2}, {"c": 3, "d": 4}], assert_equal),
    ],
)
def test_arrays(io_fixture, water_molecule, key, value, assert_fn):
    """Generic test for different calc types."""
    # Assign the value to the molecule's info
    water_molecule.arrays[key] = value

    # Append to the ZnH5MD IO object
    io_fixture.append(water_molecule)

    # Retrieve and test
    assert_fn(io_fixture[0].arrays[key], value)
    # assert key not in io_fixture[0].arrays
    assert io_fixture[0].info == {}
    assert io_fixture[0].calc is None



import ase.build
import ase.collections
import ase.io
import h5py
import numpy as np
import pytest
from ase.calculators.calculator import all_properties
from ase.calculators.singlepoint import SinglePointCalculator

import znh5md
from znh5md.serialization import Frames


def test_IO_extend(tmp_path):
    io = znh5md.IO(tmp_path / "test.h5")
    images = list(ase.collections.s22)
    io.extend(images)

    structures = io[:]
    assert len(structures) == len(images)
    for a, b in zip(images, structures):
        assert np.array_equal(a.get_atomic_numbers(), b.get_atomic_numbers())
        assert np.allclose(a.get_positions(), b.get_positions())


def test_IO_len(tmp_path, s22_info_arrays_calc):
    io = znh5md.IO(tmp_path / "test.h5")
    io.extend(s22_info_arrays_calc)
    assert len(io) == 22


def test_IO_append(tmp_path):
    io = znh5md.IO(tmp_path / "test.h5")
    images = list(ase.collections.s22)
    io.extend(images)
    io.append(images[0])

    assert len(io) == len(images) + 1
    for a, b in zip(images + [images[0]], io[:]):
        assert np.array_equal(a.get_atomic_numbers(), b.get_atomic_numbers())
        assert np.allclose(a.get_positions(), b.get_positions())


def test_author_creater(tmp_path):
    io = znh5md.IO(
        tmp_path / "test.h5",
        author="Fabian",
        author_email="email@uni-stuttgart.de",
        creator="ZnH5MD",
        creator_version="V0.3",
    )
    io.extend(list(ase.collections.s22))

    io2 = znh5md.IO(tmp_path / "test.h5")

    assert io2.author == "Fabian"
    assert io2.author_email == "email@uni-stuttgart.de"
    assert io2.creator == "ZnH5MD"
    assert io2.creator_version == "V0.3"


def test_extend_empty(tmp_path):
    io = znh5md.IO(tmp_path / "test.h5")
    io.extend(list(ase.collections.s22))

    assert len(io) == 22
    with pytest.warns(UserWarning, match="No data provided"):
        io.extend([])
    assert len(io) == 22


def test_extend_single(tmp_path):
    vectors = np.random.rand(3, 3, 2, 3)

    water = ase.build.molecule("H2O")
    water.info["vectors"] = vectors

    znh5md.write(tmp_path / "test.h5", water)

    io = znh5md.IO(tmp_path / "test.h5")
    assert len(io) == 1
    assert np.allclose(io[0].info["vectors"], vectors)
    assert io[0].info["vectors"].shape == vectors.shape


def test_not_use_ase_calc_read(tmp_path, s22_all_properties):
    io = znh5md.IO(tmp_path / "test.h5")
    io.extend(s22_all_properties)

    new_io = znh5md.IO(tmp_path / "test.h5", use_ase_calc=False)

    atoms = new_io[0]
    assert atoms.calc is None
    for key, val in s22_all_properties[0].info.items():
        assert atoms.info[key] == val

    for key, val in s22_all_properties[0].arrays.items():
        assert np.allclose(atoms.arrays[key], val)

    for key, val in s22_all_properties[0].calc.results.items():
        if key in atoms.arrays:
            assert np.allclose(atoms.arrays[key], val)
        else:
            assert np.allclose(atoms.info[key], val)


@pytest.mark.parametrize("key", all_properties + ["dummy"])
def test_not_use_ase_calc_write_arrays(tmp_path, key):
    water = ase.build.molecule("H2O")
    # When ignoring the `use_calc` the info/arrays split
    # is not dependent on the type so we use dummy data for each
    water.arrays[key] = np.random.rand(len(water), 3)
    assert key in water.arrays

    io = znh5md.IO(tmp_path / "test.h5", use_ase_calc=False)
    io.append(water)

    assert io[0].calc is None
    assert key in io[0].arrays
    assert np.allclose(io[0].arrays[key], water.arrays[key])


@pytest.mark.parametrize("key", all_properties + ["dummy"])
def test_not_use_ase_calc_write_info(tmp_path, key):
    water = ase.build.molecule("H2O")
    # When ignoring the `use_calc` the info/arrays split
    # is not dependent on the type so we use dummy data for each
    water.info[key] = np.random.rand()
    assert key in water.info

    io = znh5md.IO(tmp_path / "test.h5", use_ase_calc=False)
    io.append(water)

    assert io[0].calc is None
    assert key in io[0].info
    assert np.allclose(io[0].info[key], water.info[key])


@pytest.mark.parametrize("info_key", all_properties + ["dummy"])
@pytest.mark.parametrize("arrays_key", all_properties + ["dummy"])
def test_not_use_ase_calc_write_info_arrays(tmp_path, info_key, arrays_key):
    water = ase.build.molecule("H2O")
    # When ignoring the `use_calc` the info/arrays split
    # is not dependent on the type so we use dummy data for each
    water.info[info_key] = np.random.rand()
    water.info[f"{info_key}_arr"] = np.random.rand(
        1,
    )
    water.arrays[arrays_key] = np.random.rand(len(water), 3)
    assert info_key in water.info
    assert arrays_key in water.arrays

    io = znh5md.IO(tmp_path / "test.h5", use_ase_calc=False)
    if info_key == arrays_key:
        with pytest.raises(ValueError):
            io.append(water)

        return
    else:
        io.append(water)

    assert io[0].calc is None
    assert info_key in io[0].info
    assert arrays_key in io[0].arrays
    assert np.allclose(io[0].info[info_key], water.info[info_key])
    assert np.allclose(io[0].arrays[arrays_key], water.arrays[arrays_key])
    assert isinstance(io[0].info[info_key], float)
    assert isinstance(io[0].arrays[arrays_key], np.ndarray)
    assert isinstance(io[0].info[f"{info_key}_arr"], np.ndarray)


@pytest.mark.parametrize("key", all_properties)
def test_ase_info_key_value_error_info(tmp_path, key):
    water = ase.build.molecule("H2O")
    water.info[key] = np.random.rand()
    water.calc = SinglePointCalculator(water, **{key: np.random.rand()})
    with pytest.raises(ValueError):
        Frames.from_ase([water]).check()

    assert key in water.info

    io = znh5md.IO(tmp_path / "test.h5")
    with pytest.raises(ValueError):
        io.append(water)


@pytest.mark.parametrize("key", all_properties)
def test_ase_info_key_value_error_arrays(tmp_path, key):
    water = ase.build.molecule("H2O")
    water.arrays[key] = np.random.rand(len(water), 3)
    water.calc = SinglePointCalculator(water, **{key: np.random.rand()})
    with pytest.raises(ValueError):
        Frames.from_ase([water]).check()

    assert key in water.arrays

    io = znh5md.IO(tmp_path / "test.h5")
    with pytest.raises(ValueError):
        io.append(water)


@pytest.mark.parametrize("use_ase_calc", [True, False])
def test_convert_extxzy(tmp_path, s22_energy_forces, use_ase_calc):
    extxyz_fle = tmp_path / "test.extxyz"
    ase.io.write(extxyz_fle, s22_energy_forces)

    atoms = list(ase.io.iread(extxyz_fle))
    assert len(atoms) == len(s22_energy_forces)

    io = znh5md.IO(tmp_path / "test.h5", use_ase_calc=True)
    io.extend(atoms)

    # read with and without calc
    io = znh5md.IO(tmp_path / "test.h5", use_ase_calc=use_ase_calc)

    if use_ase_calc:
        assert "energy" in io[0].calc.results
        assert "forces" in io[0].calc.results
    else:
        assert "energy" in io[0].info
        assert "forces" in io[0].arrays


@pytest.mark.parametrize("expect_float", [True, False])
@pytest.mark.parametrize("key", all_properties)
def test_use_ase_calc_write_info_arrays(tmp_path, key, expect_float):
    water = ase.build.molecule("H2O")
    # When ignoring the `use_calc` the info/arrays split
    # is not dependent on the type so we use dummy data for each
    scalar_property = key in ["energy", "magmom", "free_energy"]
    if scalar_property:
        val = (
            np.random.rand()
            if expect_float
            else np.random.rand(
                1,
            )
        )
        calc = SinglePointCalculator(water, **{key: val})
    else:
        val = np.random.rand(len(water), 3)
        calc = SinglePointCalculator(water, **{key: val})
    water.calc = calc

    io = znh5md.IO(tmp_path / "test.h5")
    io.append(water)
    if scalar_property:
        assert np.allclose(val, calc.results[key])
        assert (
            isinstance(calc.results[key], float)
            if expect_float
            else isinstance(calc.results[key], np.ndarray)
        )
    else:
        assert np.allclose(val, calc.results[key])
        assert isinstance(calc.results[key], np.ndarray)


def test_index_error(tmp_path):
    io = znh5md.IO(tmp_path / "test.h5")
    with pytest.raises(FileNotFoundError):
        # file does not exist yet
        io[0]

    io.append(ase.build.molecule("H2O"))
    assert io[0] is not None
    assert io[-1] is not None
    assert io[0:1] is not None
    assert io[0:2] is not None
    with pytest.raises(IndexError):
        io[1]
    with pytest.raises(IndexError):
        io[-2]


@pytest.mark.skip("ASE 3.23 does not support this anymore")
def test_np_int_getitem(tmp_path):
    io = znh5md.IO(tmp_path / "test.h5")
    io.append(ase.build.molecule("H2O"))
    assert io[np.int64(0)] is not None
    assert io[np.int64(-1)] is not None
    with pytest.raises(IndexError):
        io[np.int64(1)]
    with pytest.raises(IndexError):
        io[np.int64(-2)]


@pytest.mark.parametrize("chunk_size", (None, 4, (4, 4), [4, 4, 4], (4, 4, 4, 4)))
def test_chunk_size(tmp_path, chunk_size, s22_energy_forces):
    io = znh5md.IO(tmp_path / "test.h5", chunk_size=chunk_size)
    io.extend(s22_energy_forces)
    with h5py.File(tmp_path / "test.h5", "r") as f:
        ds = f["particles/atoms/species/value"]
        assert ds.shape == (22, 30)
        if chunk_size is None:
            assert ds.chunks == (22, 30)
        elif chunk_size == 4:
            assert ds.chunks == (4, 30)
        else:
            assert ds.chunks == (4, 4)
    with h5py.File(tmp_path / "test.h5", "r") as f:
        ds = f["particles/atoms/force/value"]
        assert ds.shape == (22, 30, 3)
        if chunk_size is None:
            assert ds.chunks == (11, 30, 3)  # auto chunking
        elif chunk_size == 4:
            assert ds.chunks == (4, 30, 3)
        else:
            assert ds.chunks == (4, 4, 3)



from znh5md.path import get_h5md_path
from znh5md.serialization import Frames


def test_get_h5md_path(
    s22_all_properties, s22_info_arrays_calc, frames_with_residuenames
):
    frame = Frames.from_ase(
        s22_all_properties + s22_info_arrays_calc + frames_with_residuenames
    )

    assert get_h5md_path("positions", "atoms", frame) == "/particles/atoms/position"
    assert (
        get_h5md_path("energy", "atoms", frame) == "/observables/atoms/potential_energy"
    )
    assert get_h5md_path("forces", "atoms", frame) == "/particles/atoms/force"

    assert get_h5md_path("cell", "atoms", frame) == "/particles/atoms/box/edges"
    assert get_h5md_path("pbc", "atoms", frame) == "/particles/atoms/box/pbc"

    assert (
        get_h5md_path("mlip_forces", "atoms", frame) == "/particles/atoms/mlip_forces"
    )
    assert (
        get_h5md_path("mlip_forces_2", "atoms", frame)
        == "/particles/atoms/mlip_forces_2"
    )
    # assert get_h5md_path("momenta", "atoms", frame) == "/particles/atoms/momenta"
    # assert get_h5md_path("velocity", "atoms", frame) == "/particles/atoms/momenta"
    assert (
        get_h5md_path("residuenames", "atoms", frame) == "/particles/atoms/residuenames"
    )
    assert get_h5md_path("atomtypes", "atoms", frame) == "/particles/atoms/atomtypes"

    assert (
        get_h5md_path("mlip_energy", "atoms", frame) == "/observables/atoms/mlip_energy"
    )
    assert (
        get_h5md_path("mlip_energy_2", "atoms", frame)
        == "/observables/atoms/mlip_energy_2"
    )
    assert (
        get_h5md_path("mlip_stress", "atoms", frame) == "/observables/atoms/mlip_stress"
    )
    assert (
        get_h5md_path("collection", "atoms", frame) == "/observables/atoms/collection"
    )
    assert get_h5md_path("metadata", "atoms", frame) == "/observables/atoms/metadata"

    assert get_h5md_path("energies", "atoms", frame) == "/particles/atoms/energies"
    assert (
        get_h5md_path("free_energy", "atoms", frame) == "/observables/atoms/free_energy"
    )
    assert get_h5md_path("stress", "atoms", frame) == "/observables/atoms/stress"
    assert get_h5md_path("stresses", "atoms", frame) == "/particles/atoms/stresses"
    assert get_h5md_path("dipole", "atoms", frame) == "/observables/atoms/dipole"
    assert get_h5md_path("magmom", "atoms", frame) == "/observables/atoms/magmom"
    assert get_h5md_path("magmoms", "atoms", frame) == "/particles/atoms/magmoms"
    assert (
        get_h5md_path("dielectric_tensor", "atoms", frame)
        == "/observables/atoms/dielectric_tensor"
    )
    assert (
        get_h5md_path("born_effective_charges", "atoms", frame)
        == "/particles/atoms/born_effective_charges"
    )
    assert (
        get_h5md_path("polarization", "atoms", frame)
        == "/observables/atoms/polarization"
    )



import h5py
import numpy as np
import numpy.testing as npt
import pytest
import rdkit2ase

from znh5md.interface import IO


@pytest.mark.parametrize(
    "dataset_name",
    [
        "s22",
        "s22_energy",
        "s22_energy_forces",
        "s22_all_properties",
        "s22_info_arrays_calc",
        "s22_mixed_pbc_cell",
        "s22_illegal_calc_results",
        "s22_no_ascii",
        "frames_with_residuenames",
        "s22_info_arrays_calc_missing_inbetween",
    ],
)
@pytest.mark.parametrize("append", [False])  # TODO: True raises error!!
# @pytest.mark.parametrize("store_ase_origin ", [False, True])
def test_frames_iter(dataset_name, append, request, tmp_path):
    dataset = request.getfixturevalue(dataset_name)

    io = IO(tmp_path / "test.h5")
    # io._store_ase_origin = store_ase_origin
    if append:
        for frame in dataset:
            io.append(frame)
    else:
        io.extend(dataset)

    assert len(io) == len(dataset)
    for a, b in zip(io, dataset):
        assert a == b
        for key in set(a.arrays.keys()) | set(b.arrays.keys()):
            npt.assert_array_equal(a.arrays[key], b.arrays[key])
        for key in set(a.info.keys()) | set(b.info.keys()):
            npt.assert_array_equal(a.info[key], b.info[key])
        if b.calc is not None or a.calc is not None:
            for key in set(a.calc.results.keys()) | set(b.calc.results.keys()):
                npt.assert_array_equal(a.calc.results[key], b.calc.results[key])


def test_arrays_list_array(tmp_path):
    atoms = rdkit2ase.smiles2atoms("Cl")
    atoms.arrays["lst"] = list(atoms.symbols)
    atoms.arrays["arr"] = np.array(atoms.symbols)

    io = IO(tmp_path / "test.h5")
    io.extend([atoms, rdkit2ase.smiles2atoms("C")])

    with h5py.File(io.filename, "r") as f:
        assert np.array_equal(f["particles/atoms/lst/value"][0], b'["Cl", "H"]')
        assert np.array_equal(f["particles/atoms/lst/value"][1], b"")
        # now for arr
        assert np.array_equal(f["particles/atoms/arr/value"][0], b'["Cl", "H"]')
        assert np.array_equal(f["particles/atoms/arr/value"][1], b"")



import numpy as np
import numpy.testing as npt
import rdkit2ase

import znh5md


def test_pdb_repeat(tmp_path):
    water = rdkit2ase.smiles2conformers(smiles="O", numConfs=1)[0]
    water.arrays["atomtypes"] = np.array(["H", "O", "H"])
    assert isinstance(water.arrays["atomtypes"], np.ndarray)

    io = znh5md.IO(tmp_path / "test.h5")
    io.append(water)

    atoms = io[0]
    assert isinstance(atoms.arrays["atomtypes"], np.ndarray)
    npt.assert_array_equal(atoms.arrays["atomtypes"], water.arrays["atomtypes"])



import contextlib
import typing as t

import ase.io
import h5py
import numpy.testing as npt
import pytest

import znh5md


@pytest.mark.parametrize(
    "dataset",
    [
        "s22",
        "s22_energy",
        "s22_all_properties",
        "s22_info_arrays_calc",
        "s22_mixed_pbc_cell",
        "s22_illegal_calc_results",
        "water",
        "s22_nested_calc",
    ],
)
def test_datasets(tmp_path, dataset, request):
    images = request.getfixturevalue(dataset)
    io = znh5md.IO(tmp_path / "test.h5")
    io.extend(images)
    images2 = io[:]

    assert len(images) == len(images2)

    for a, b in zip(images, images2):
        npt.assert_array_equal(a.get_positions(), b.get_positions())
        npt.assert_array_equal(a.get_atomic_numbers(), b.get_atomic_numbers())
        npt.assert_array_equal(a.get_cell(), b.get_cell())
        npt.assert_array_equal(a.get_pbc(), b.get_pbc())
        npt.assert_array_equal(a.get_velocities(), b.get_velocities())
        if a.calc is not None:
            assert set(a.calc.results) == set(b.calc.results)
            for key in a.calc.results:
                npt.assert_array_equal(a.calc.results[key], b.calc.results[key])
            if "energy" in a.calc.results:
                assert b.get_potential_energy() == a.get_potential_energy()
                assert isinstance(a.get_potential_energy(), float)
                assert isinstance(b.get_potential_energy(), float)
        else:
            assert b.calc is None

        assert set(a.arrays) == set(b.arrays)
        for key in a.arrays:
            npt.assert_array_equal(a.arrays[key], b.arrays[key])

        # h5md keys are added to info automatically
        assert set(b.info) == set(a.info)
        for key in a.info:
            npt.assert_array_equal(a.info[key], b.info[key])


@pytest.mark.parametrize(
    "dataset",
    [
        "s22",
        "s22_energy",
        "s22_all_properties",
        "s22_info_arrays_calc",
        "s22_mixed_pbc_cell",
        "water",
    ],
)
def test_datasets_extxyz(tmp_path, dataset, request):
    images = request.getfixturevalue(dataset)
    ase.io.write(tmp_path / "test.xyz", images)
    io = znh5md.IO(tmp_path / "test.h5")
    io.extend(list(ase.io.iread(tmp_path / "test.xyz")))
    images2 = io[:]

    assert len(images) == len(images2)

    for a, b in zip(images, images2):
        npt.assert_array_almost_equal(a.get_positions(), b.get_positions())
        npt.assert_array_almost_equal(a.get_atomic_numbers(), b.get_atomic_numbers())
        npt.assert_array_almost_equal(a.get_cell(), b.get_cell())
        npt.assert_array_almost_equal(a.get_pbc(), b.get_pbc())
        npt.assert_array_almost_equal(a.get_velocities(), b.get_velocities())
        if a.calc is not None:
            assert set(a.calc.results) == set(b.calc.results)
            for key in a.calc.results:
                npt.assert_array_almost_equal(a.calc.results[key], b.calc.results[key])
            if "energy" in a.calc.results:
                assert b.get_potential_energy() == a.get_potential_energy()
                assert isinstance(a.get_potential_energy(), float)
                assert isinstance(b.get_potential_energy(), float)
        else:
            assert b.calc is None

        assert set(a.arrays) == set(b.arrays)
        for key in a.arrays:
            npt.assert_array_almost_equal(a.arrays[key], b.arrays[key])

        # h5md keys are added to info automatically
        assert set(b.info) == set(a.info)
        for key in a.info:
            if isinstance(a.info[key], str):
                assert a.info[key] == b.info[key]
            elif isinstance(a.info[key], dict):
                assert a.info[key] == b.info[key]
            else:
                npt.assert_array_almost_equal(a.info[key], b.info[key])


@pytest.mark.parametrize("store", ["linear"])
@pytest.mark.parametrize(
    "dataset",
    [
        "s22_info_arrays_calc",
    ],
)
def test_datasets_h5py(tmp_path, dataset, request, store):
    images = request.getfixturevalue(dataset)
    io = znh5md.IO(tmp_path / "test.h5", store=store, timestep=0.5)
    io.extend(images)

    with h5py.File(tmp_path / "test.h5", "r") as f:
        assert "particles/atoms/position/value" in f
        assert "particles/atoms/species/value" in f
        assert "particles/atoms/force/value" in f
        assert "observables/atoms/force/value" not in f
        assert "particles/atoms/velocity/value" in f
        assert "observables/atoms/velocity/value" not in f

        # assert "particles/atoms/momenta/value" in f
        # assert "observables/atoms/momenta/value" not in f

        assert "particles/atoms/potential_energy/value" not in f
        assert "observables/atoms/potential_energy/value" in f

        assert "particles/atoms/mlip_forces/value" in f
        assert "particles/atoms/mlip_forces_2/value" in f

        assert "observables/atoms/mlip_energy/value" in f
        assert "observables/atoms/mlip_energy_2/value" in f
        assert "observables/atoms/mlip_stress/value" in f

        assert f["particles/atoms/velocity/value"].attrs["unit"] == "Angstrom/fs"
        assert f["particles/atoms/force/value"].attrs["unit"] == "eV/Angstrom"

        # if store == "time":
        #     assert f["observables/atoms/energy/time"].shape == (len(images),)
        #     assert f["observables/atoms/energy/step"].shape == (len(images),)
        if store == "linear":
            assert f["observables/atoms/potential_energy/time"][()] == 0.5
            assert f["observables/atoms/potential_energy/step"][()] == 1

        # assert f["observables/atoms/energy/value"].attrs["unit"] == "eV"

        npt.assert_array_equal(
            f["particles/atoms/box"].attrs["boundary"], ["none", "none", "none"]
        )


@pytest.mark.parametrize("store", ["linear"])
def test_two_datasets(tmp_path, s22_all_properties, s22_mixed_pbc_cell, store):
    io_a = znh5md.IO(tmp_path / "test.h5", particles_group="a", store=store)
    io_b = znh5md.IO(tmp_path / "test.h5", particles_group="b")
    io_a.extend(s22_all_properties)
    io_b.extend(s22_mixed_pbc_cell)

    with h5py.File(tmp_path / "test.h5", "r") as f:
        assert "/particles/a/position/value" in f
        assert "/particles/b/position/value" in f

    for a, b in zip(s22_all_properties, io_a[:]):
        npt.assert_array_equal(a.get_positions(), b.get_positions())

    for a, b in zip(s22_mixed_pbc_cell, io_b[:]):
        npt.assert_array_equal(a.get_positions(), b.get_positions())


@pytest.mark.parametrize("store", ["linear"])
def test_two_datasets_file_handle(
    tmp_path, s22_all_properties, s22_mixed_pbc_cell, store
):
    with h5py.File(tmp_path / "test.h5", "w") as f:
        io_a = znh5md.IO(file_handle=f, particles_group="a")
        io_b = znh5md.IO(file_handle=f, particles_group="b", store=store)

        io_a.extend(s22_all_properties)
        io_b.extend(s22_mixed_pbc_cell)

        assert len(io_a) == len(s22_all_properties)
        assert len(io_b) == len(s22_mixed_pbc_cell)

    with h5py.File(tmp_path / "test.h5", "r") as f:
        io_a = znh5md.IO(file_handle=f, particles_group="a")
        io_b = znh5md.IO(file_handle=f, particles_group="b")

        assert len(io_a) == len(s22_all_properties)
        assert len(io_b) == len(s22_mixed_pbc_cell)


@pytest.mark.parametrize("store", ["linear"])
def test_two_datasets_file_factory(
    tmp_path, s22_all_properties, s22_mixed_pbc_cell, store
):
    io_a = znh5md.IO(filename=tmp_path / "test.h5", particles_group="a")
    io_b = znh5md.IO(filename=tmp_path / "test.h5", particles_group="b", store=store)

    io_a.extend(s22_all_properties)
    io_b.extend(s22_mixed_pbc_cell)

    assert len(io_a) == len(s22_all_properties)
    assert len(io_b) == len(s22_mixed_pbc_cell)

    def make_h5py_opener_a(filename: str):
        def _opener():
            return h5py.File(filename, mode="r")

        return _opener

    # The file_factory is read-only!
    file_factory = make_h5py_opener_a(tmp_path / "test.h5")

    io_a = znh5md.IO(file_factory=file_factory, particles_group="a")
    io_b = znh5md.IO(file_factory=file_factory, particles_group="b")

    assert len(io_a) == len(s22_all_properties)
    assert len(io_b) == len(s22_mixed_pbc_cell)

    # test another file opener factory
    def make_h5py_opener_b(
        filename: str, **h5py_kwargs
    ) -> t.Callable[[], t.ContextManager[h5py.File]]:
        @contextlib.contextmanager
        def _opener() -> t.Generator[h5py.File, None, None]:
            with open(filename, "rb") as f:
                with h5py.File(f, "r", **h5py_kwargs) as h5f:
                    yield h5f

        return _opener

    # The file_factory is read-only!
    file_factory = make_h5py_opener_b(tmp_path / "test.h5")

    io_a = znh5md.IO(file_factory=file_factory, particles_group="a")
    io_b = znh5md.IO(file_factory=file_factory, particles_group="b")

    assert len(io_a) == len(s22_all_properties)
    assert len(io_b) == len(s22_mixed_pbc_cell)


def test_pbc(tmp_path, s22_mixed_pbc_cell):
    io = znh5md.IO(tmp_path / "test.h5")
    io.extend(s22_mixed_pbc_cell)

    pbc = ["periodic" if x else "none" for x in io[0].pbc]

    with h5py.File(tmp_path / "test.h5", "r") as f:
        assert "/particles/atoms/position/value" in f
        assert "/particles/atoms/box" in f

        npt.assert_array_equal(f["particles/atoms/box"].attrs["boundary"], pbc)
        for idx, atom in enumerate(s22_mixed_pbc_cell):
            npt.assert_array_equal(
                f["particles/atoms/box/edges/value"][idx], atom.get_cell()
            )
            # Do we want to rename "pbc" to "boundary" and make if "none"
            # or "periodic" as well? This should only exist if requested
            npt.assert_array_equal(f["particles/atoms/box/pbc/value"][idx], atom.pbc)


def test_save_load_save_load(tmp_path, s22_mixed_pbc_cell):
    io1 = znh5md.IO(tmp_path / "test_1.h5")
    io1.extend(s22_mixed_pbc_cell)

    images = io1[:]
    io2 = znh5md.IO(tmp_path / "test_2.h5")
    io2.extend(images)

    images2 = io2[:]

    assert len(images) == len(images2)


@pytest.mark.parametrize("timestep", [0.5, 1.0])
def test_time_step(tmp_path, s22_mixed_pbc_cell, timestep):
    io = znh5md.IO(tmp_path / "test.h5", timestep=timestep)
    io.extend(s22_mixed_pbc_cell)
    # Do it twice, to check for appending
    io.extend(s22_mixed_pbc_cell)
    images = io[:]
    assert len(images) == len(s22_mixed_pbc_cell) * 2
    # for idx, atoms in enumerate(images):
    #     assert atoms.info["h5md_step"] == idx
    #     assert atoms.info["h5md_time"] == idx * timestep


def test_slicing(tmp_path, s22_mixed_pbc_cell):
    # Create an instance of znh5md.IO and extend it with the test data
    io = znh5md.IO(tmp_path / "test.h5")
    io.extend(s22_mixed_pbc_cell)

    # Ensure the length of the io object matches the input data
    assert len(io) == len(s22_mixed_pbc_cell)

    # Test various slicing cases

    # Single element access
    assert io[0] == s22_mixed_pbc_cell[0]
    assert io[len(io) - 1] == s22_mixed_pbc_cell[len(s22_mixed_pbc_cell) - 1]

    # Simple slices
    assert io[:10] == s22_mixed_pbc_cell[:10]
    assert io[10:20] == s22_mixed_pbc_cell[10:20]
    assert io[-10:] == s22_mixed_pbc_cell[-10:]
    assert io[:-10] == s22_mixed_pbc_cell[:-10]

    # Step slices
    assert io[::2] == s22_mixed_pbc_cell[::2]
    assert io[1::2] == s22_mixed_pbc_cell[1::2]
    # not allowed in h5py
    # assert io[::-1] == s22_mixed_pbc_cell[::-1]  # Reverse

    # Complex slices
    assert io[5:20:3] == s22_mixed_pbc_cell[5:20:3]
    assert io[-20:-5:2] == s22_mixed_pbc_cell[-20:-5:2]

    # Empty slices
    assert io[5:5] == s22_mixed_pbc_cell[5:5]
    assert io[-5:-5] == s22_mixed_pbc_cell[-5:-5]

    # Ensure slicing does not affect original length
    assert len(io) == len(s22_mixed_pbc_cell)



import ase.build
import networkx as nx
import numpy as np
import pytest
from ase.neighborlist import natural_cutoffs


def build_graph(atoms: ase.Atoms) -> nx.Graph:
    cutoffs = [
        1.0,
    ]
    atoms_copy = atoms.copy()
    connectivity_matrix = np.zeros((len(atoms_copy), len(atoms_copy)), dtype=int)
    atoms_copy.pbc = False
    distance_matrix = atoms_copy.get_all_distances(mic=False)
    np.fill_diagonal(distance_matrix, np.inf)
    for cutoff in cutoffs:
        cutoffs = np.array(natural_cutoffs(atoms_copy, mult=cutoff))
        cutoffs = cutoffs[:, None] + cutoffs[None, :]
        connectivity_matrix[distance_matrix <= cutoffs] += 1
    return nx.from_numpy_array(connectivity_matrix)


def set_connectivity(atoms: ase.Atoms, graph: nx.Graph):
    bonds = []
    for edge in graph.edges:
        bonds.append((edge[0], edge[1], graph.edges[edge]["weight"]))
    atoms.connectivity = bonds


@pytest.fixture
def connected_atoms() -> list[ase.Atoms]:
    """Create ase atoms that are connected."""
    water = ase.build.molecule("H2O")
    ammonia = ase.build.molecule("NH3")
    methane = ase.build.molecule("CH4")

    set_connectivity(water, build_graph(water))
    set_connectivity(ammonia, build_graph(ammonia))
    set_connectivity(methane, build_graph(methane))

    return [water, ammonia, methane]


def test_connectivity(tmp_path, connected_atoms):
    assert connected_atoms[0].connectivity == [(0, 1, 1), (0, 2, 1)]



import json

import h5py
import numpy as np
import numpy.testing as npt

from znh5md.misc import MISSING
from znh5md.serialization import Entry


def test_entry_list_int():
    e = Entry(value=[1, 2, 3], origin=None, name="test")
    assert e.ref == 1
    assert np.isnan(e.fillvalue)
    val, dtype = e.dump()
    assert isinstance(val, np.ndarray)
    npt.assert_array_equal(val, [1, 2, 3])
    assert dtype == np.float64


def test_entry_array_int():
    e = Entry(value=np.array([1, 2, 3]), origin=None, name="test")
    assert e.ref == 1
    assert np.isnan(e.fillvalue)
    val, dtype = e.dump()
    assert isinstance(val, np.ndarray)
    npt.assert_array_equal(val, [1, 2, 3])
    assert val.dtype == np.float64
    assert dtype == np.float64


def test_entry_list_int_missing():
    e = Entry(value=[1, MISSING, 3], origin=None, name="test")
    assert e.ref == 1
    assert np.isnan(e.fillvalue)
    val, dtype = e.dump()
    assert dtype == np.float64
    assert isinstance(val, np.ndarray)
    npt.assert_array_equal(val, [1, np.nan, 3])


def test_entry_array_int_missing():
    e = Entry(value=np.array([1, MISSING, 3]), origin=None, name="test")
    assert e.ref == 1
    assert np.isnan(e.fillvalue)
    val, dtype = e.dump()
    assert dtype == np.float64
    assert isinstance(val, np.ndarray)
    npt.assert_array_equal(val, [1, np.nan, 3])


def test_entry_int_missing_first():
    e = Entry(value=np.array([MISSING, 2]), origin=None, name="test")
    assert e.ref == 2
    assert np.isnan(e.fillvalue)
    val, dtype = e.dump()
    assert dtype == np.float64
    assert isinstance(val, np.ndarray)
    npt.assert_array_equal(val, [np.nan, 2])


def test_entry_list_str():
    e = Entry(value=["a", "b", "c"], origin=None, name="test")
    assert e.ref == "a"
    assert e.fillvalue == ""

    val, dtype = e.dump()
    assert isinstance(val, list)
    assert val == [json.dumps(x) for x in ["a", "b", "c"]]
    assert dtype == h5py.string_dtype()


def test_entry_array_str():
    e = Entry(value=np.array(["a", "b", "c"]), origin=None, name="test")
    assert e.ref == "a"
    assert e.fillvalue == ""

    val, dtype = e.dump()
    assert isinstance(val, list)
    assert val == [json.dumps(x) for x in ["a", "b", "c"]]
    assert dtype == h5py.string_dtype()


def test_entry_list_str_missing():
    e = Entry(value=["a", MISSING, "c"], origin=None, name="test")
    assert e.ref == "a"
    assert e.fillvalue == ""

    val, dtype = e.dump()
    assert isinstance(val, list)
    assert val == [json.dumps("a"), "", json.dumps("c")]
    assert dtype == h5py.string_dtype()


def test_entry_array_str_missing():
    e = Entry(value=np.array(["a", MISSING, "c"]), origin=None, name="test")
    assert e.ref == "a"
    assert e.fillvalue == ""

    val, dtype = e.dump()
    assert isinstance(val, list)
    assert val == [json.dumps("a"), "", json.dumps("c")]
    assert dtype == h5py.string_dtype()


def test_entry_list_dict():
    e = Entry(value=[{"a": 1}, {"b": 2}], origin=None, name="test")
    assert e.ref == {"a": 1}
    assert e.fillvalue == ""

    val, dtype = e.dump()
    assert isinstance(val, list)
    assert val == [json.dumps(x) for x in [{"a": 1}, {"b": 2}]]
    assert dtype == h5py.string_dtype()


def test_entry_array_dict():
    e = Entry(value=np.array([{"a": 1}, {"b": 2}]), origin=None, name="test")
    assert e.ref == {"a": 1}
    assert e.fillvalue == ""

    val, dtype = e.dump()
    assert isinstance(val, list)
    assert val == [json.dumps(x) for x in [{"a": 1}, {"b": 2}]]
    assert dtype == h5py.string_dtype()


def test_entry_list_dict_missing():
    e = Entry(value=[{"a": 1}, MISSING, {"b": 2}], origin=None, name="test")
    assert e.ref == {"a": 1}
    assert e.fillvalue == ""

    val, dtype = e.dump()
    assert isinstance(val, list)
    assert val == [json.dumps({"a": 1}), "", json.dumps({"b": 2})]
    assert dtype == h5py.string_dtype()


def test_entry_array_dict_missing():
    e = Entry(value=np.array([MISSING, {"a": 1}, {"b": 2}]), origin=None, name="test")
    assert e.ref == {"a": 1}
    assert e.fillvalue == ""

    val, dtype = e.dump()
    assert isinstance(val, list)
    assert val == ["", json.dumps({"a": 1}), json.dumps({"b": 2})]
    assert dtype == h5py.string_dtype()


def test_entry_list_array_int():
    e = Entry(value=[np.array([1, 2]), np.array([3, 4])], origin=None, name="test")
    npt.assert_array_equal(e.ref, np.array([1, 2]))
    assert np.isnan(e.fillvalue)

    val, dtype = e.dump()
    assert isinstance(val, np.ndarray)
    npt.assert_array_equal(val, [[1, 2], [3, 4]])
    assert dtype == np.float64


def test_entry_list_array_int_missing():
    e = Entry(
        value=[np.array([1, 2]), MISSING, np.array([3, 4])], origin=None, name="test"
    )
    npt.assert_array_equal(e.ref, np.array([1, 2]))
    assert np.isnan(e.fillvalue)

    val, dtype = e.dump()
    assert isinstance(val, np.ndarray)
    npt.assert_array_equal(val, [[1, 2], [np.nan, np.nan], [3, 4]])
    assert dtype == np.float64


def test_entry_list_array_str():
    e = Entry(
        value=[np.array(["a", "b"]), np.array(["c", "d"])], origin=None, name="test"
    )
    npt.assert_array_equal(e.ref, np.array(["a", "b"]))
    assert e.fillvalue == ""

    val, dtype = e.dump()
    assert isinstance(val, list)
    assert val == [json.dumps(["a", "b"]), json.dumps(["c", "d"])]
    assert dtype == h5py.string_dtype()


def test_entry_list_array_str_missing():
    e = Entry(
        value=[np.array(["a", "b"]), MISSING, np.array(["c", "d"])],
        origin=None,
        name="test",
    )
    npt.assert_array_equal(e.ref, np.array(["a", "b"]))
    assert e.fillvalue == ""

    val, dtype = e.dump()
    assert isinstance(val, list)
    assert val == [json.dumps(["a", "b"]), "", json.dumps(["c", "d"])]
    assert dtype == h5py.string_dtype()


def test_list_list_str_missing():
    e = Entry(value=[["Cl", "H"], ["O", "O"], MISSING], origin=None, name="test")
    npt.assert_array_equal(e.ref, np.array(["Cl", "H"]))
    assert e.fillvalue == ""

    val, dtype = e.dump()
    assert isinstance(val, list)
    assert val == [json.dumps(["Cl", "H"]), json.dumps(["O", "O"]), ""]
    assert dtype == h5py.string_dtype()


def test_list_list_str_special():
    e = Entry(value=[["Cl", "H"], ["O", "O"], None, ""], origin=None, name="test")
    npt.assert_array_equal(e.ref, np.array(["Cl", "H"]))
    assert e.fillvalue == ""

    val, dtype = e.dump()
    assert isinstance(val, list)
    assert val == [json.dumps(["Cl", "H"]), json.dumps(["O", "O"]), "null", '""']
    assert dtype == h5py.string_dtype()


def test_list_list_ndarray():
    e = Entry(value=[[np.array([1, 2]), np.array([3, 4])]], origin=None, name="test")
    npt.assert_array_equal(e.ref, np.array([np.array([1, 2]), np.array([3, 4])]))
    assert np.isnan(e.fillvalue)

    val, dtype = e.dump()
    assert isinstance(val, np.ndarray)
    npt.assert_array_equal(val[0], [[1, 2], [3, 4]])
    assert dtype == np.float64



"""Test the CLI module."""

import ase.build
import ase.io
from typer.testing import CliRunner

import znh5md
from znh5md.cli import app  # Import the Typer app

runner = CliRunner()


def test_export(tmp_path):
    tmp_input = tmp_path / "input.h5md"
    tmp_output = tmp_path / "output.xyz"

    atoms = ase.build.molecule("H2O")
    znh5md.write(tmp_input, atoms)

    assert tmp_input.exists()

    result = runner.invoke(app, ["export", tmp_input.as_posix(), tmp_output.as_posix()])
    assert result.exit_code == 0
    assert tmp_output.exists()

    new_atoms = ase.io.read(tmp_output)
    assert new_atoms.get_chemical_symbols() == atoms.get_chemical_symbols()


def test_convert(tmp_path):
    tmp_input = tmp_path / "input.xyz"
    tmp_output = tmp_path / "output.h5md"

    atoms = ase.build.molecule("H2O")
    ase.io.write(tmp_input, atoms)

    assert tmp_input.exists()

    result = runner.invoke(
        app, ["convert", tmp_input.as_posix(), tmp_output.as_posix()]
    )
    assert result.exit_code == 0
    converted_atoms = znh5md.IO(tmp_output)[:]
    assert len(converted_atoms) > 0


def test_version():
    result = runner.invoke(app, ["--version"])
    assert result.exit_code == 0
    assert "znh5md" in result.stdout



import ase
import numpy.testing as npt
from ase.calculators.singlepoint import SinglePointCalculator

import znh5md


def test_empty(tmp_path):
    atoms = ase.Atoms()

    io = znh5md.IO(tmp_path / "frames.h5")
    io.append(atoms)

    new_atoms = io[0]
    assert len(new_atoms) == len(atoms)
    assert new_atoms.info.keys() == atoms.info.keys()
    for key in atoms.info.keys():
        npt.assert_equal(new_atoms.info[key], atoms.info[key])
    assert new_atoms.arrays.keys() == atoms.arrays.keys()
    for key in atoms.arrays.keys():
        npt.assert_equal(new_atoms.arrays[key], atoms.arrays[key])
    assert new_atoms.calc is None


def test_empty_info(tmp_path):
    atoms = ase.Atoms()
    atoms.info["test"] = []

    io = znh5md.IO(tmp_path / "frames.h5")
    io.append(atoms)

    new_atoms = io[0]
    assert len(new_atoms) == len(atoms)
    # ZnH5MD will remove keys in info which are an empty list!!

    # assert new_atoms.info.keys() == atoms.info.keys()
    # for key in atoms.info.keys():
    #     npt.assert_equal(new_atoms.info[key], atoms.info[key])
    assert new_atoms.arrays.keys() == atoms.arrays.keys()
    for key in atoms.arrays.keys():
        npt.assert_equal(new_atoms.arrays[key], atoms.arrays[key])
    assert new_atoms.calc is None


def test_empty_array(tmp_path):
    atoms = ase.Atoms()
    atoms.arrays["test"] = []

    io = znh5md.IO(tmp_path / "frames.h5")
    io.append(atoms)

    new_atoms = io[0]
    assert len(new_atoms) == len(atoms)
    assert new_atoms.info.keys() == atoms.info.keys()
    for key in atoms.info.keys():
        npt.assert_equal(new_atoms.info[key], atoms.info[key])

    # ZnH5MD will remove keys in arrays which are an empty list!!

    # assert new_atoms.arrays.keys() == atoms.arrays.keys()
    # for key in atoms.arrays.keys():
    #     npt.assert_equal(new_atoms.arrays[key], atoms.arrays[key])
    assert new_atoms.calc is None


def test_empty_calc(tmp_path):
    atoms = ase.Atoms()
    atoms.calc = SinglePointCalculator(atoms, energy=0.0)
    atoms.calc.results["test"] = []
    io = znh5md.IO(tmp_path / "frames.h5")
    io.append(atoms)

    new_atoms = io[0]
    assert len(new_atoms) == len(atoms)
    assert new_atoms.info.keys() == atoms.info.keys()
    for key in atoms.info.keys():
        npt.assert_equal(new_atoms.info[key], atoms.info[key])
    assert new_atoms.arrays.keys() == atoms.arrays.keys()
    for key in atoms.arrays.keys():
        npt.assert_equal(new_atoms.arrays[key], atoms.arrays[key])

    # ZnH5MD will remove keys in calc.results which are an empty list!!

    # assert new_atoms.calc.results.keys() == atoms.calc.results.keys()
    # for key in atoms.calc.results.keys():
    #     npt.assert_equal(new_atoms.calc.results[key], atoms.calc.results[key])



import typing as t

import numpy as np
import numpy.testing as npt
import pytest

import znh5md


def test_keys_missing(tmp_path, s22, s22_energy_forces):
    io = znh5md.IO(tmp_path / "test.h5")

    images = s22_energy_forces + s22
    io.extend(images)
    assert len(io) == len(images)
    assert len(list(io)) == len(images)

    for a, b in zip(images, io):
        assert a == b
        if b.calc is not None:
            for key in b.calc.results:
                npt.assert_array_equal(a.calc.results[key], b.calc.results[key])
        else:
            assert a.calc is None


@pytest.mark.parametrize("state", ["before", "middle", "after"])
def test_velocity(tmp_path, s22, state: t.Literal["before", "middle", "after"]):
    io = znh5md.IO(tmp_path / "test.h5")

    velocity = None

    images = s22
    if state == "before":
        velocity = np.random.random((len(images[0]), 3)) * 0.1
        images[0].set_velocities(velocity)
    elif state == "middle":
        velocity = np.random.random((len(images[1]), 3)) * 0.1
        images[1].set_velocities(velocity)
    elif state == "after":
        velocity = np.random.random((len(images[-1]), 3)) * 0.1
        images[-1].set_velocities(velocity)

    for atoms in images:
        io.append(atoms)
    assert len(io) == len(images)
    assert len(list(io)) == len(images)
    if state == "before":
        npt.assert_array_almost_equal(io[0].get_velocities(), velocity)
        npt.assert_array_almost_equal(io[:][0].get_velocities(), velocity)
    elif state == "middle":
        npt.assert_array_almost_equal(io[1].get_velocities(), velocity)
        npt.assert_array_almost_equal(io[:][1].get_velocities(), velocity)
    elif state == "after":
        npt.assert_array_almost_equal(io[-1].get_velocities(), velocity)
        npt.assert_array_almost_equal(io[:][-1].get_velocities(), velocity)



import ase.io
from ase.io.proteindatabank import read_proteindatabank, write_proteindatabank

import znh5md

from .abc import IOBase


class ASEIO(IOBase):
    def setup(self):
        pass

    def read(self) -> list[ase.Atoms]:
        if self.format == "pdb":
            return read_proteindatabank(self.filename, index=slice(None, None, None))
        elif self.format == "h5md":
            # TODO
            return list(znh5md.IO(self.filename))  # basically what ase does
        elif self.format == "xtc":
            raise ValueError("xtc format not supported")
        else:
            return ase.io.read(self.filename, format=self.format, index=":")
        # return list(ase.io.iread(self.filename, format=self.format))
        # no performance difference

    def write(self, atoms: list[ase.Atoms]) -> None:
        if self.format == "pdb":
            write_proteindatabank(self.filename, atoms)
        elif self.format == "h5md":
            znh5md.IO(self.filename, store="time").extend(atoms)
        elif self.format == "xtc":
            raise ValueError("xtc format not supported")
        else:
            ase.io.write(self.filename, atoms, format=self.format)


class ASECreate(IOBase):
    def setup(self):
        if self.format == "pdb":
            self.frames = read_proteindatabank(
                self.filename, index=slice(None, None, None)
            )
        elif self.format == "h5md":
            self.frames = list(znh5md.IO(self.filename))
        elif self.format == "xtc":
            self.frames = []  # TODO!
        else:
            self.frames = list(ase.io.iread(self.filename, format=self.format))

    def read(self) -> list[ase.Atoms]:
        frames = []
        for atoms in self.frames:
            frames.append(
                ase.Atoms(
                    positions=atoms.get_positions(),
                )
            )

    def write(self, atoms: list[ase.Atoms]) -> None:
        raise NotImplementedError



import numpy.testing as npt
import pytest

import znh5md


@pytest.mark.parametrize("index", ["slice", "list", "int", "iter"])
def test_masked(tmp_path, full_water, index):
    """Test the variable shape off functionality."""
    io = znh5md.IO(tmp_path / "test.h5", variable_shape=False)
    io.extend([full_water for _ in range(10)])

    io_masked = znh5md.IO(
        tmp_path / "test.h5",
        variable_shape=False,
        mask=[0, 2],
    )

    if index == "slice":
        io_masked = io_masked[:]
        assert len(io_masked) == 10
    elif index == "list":
        io_masked = io_masked[[0, 2]]
        assert len(io_masked) == 2
    elif index == "int":
        io_masked = [io_masked[0]]
        assert len(io_masked) == 1
    elif index == "iter":
        assert len(io_masked) == 10

    for atoms in io_masked:
        assert len(atoms) == 2
        for key, val in atoms.arrays.items():
            npt.assert_equal(
                val,
                full_water[[0, 2]].arrays[key],
            )
        for key, val in atoms.info.items():
            npt.assert_equal(
                val,
                full_water[[0, 2]].info[key],
            )
        for key, val in atoms.calc.results.items():
            if isinstance(val, (int, float)):
                assert val == full_water.calc.results[key]
            else:
                npt.assert_equal(
                    val,
                    full_water.calc.results[key][[0, 2]],
                )


def test_masked_variable_shape():
    with pytest.raises(ValueError):
        znh5md.IO(
            "test.h5",
            variable_shape=True,
            mask=[0, 2],
        )


@pytest.mark.parametrize("index", ["slice", "list", "int", "iter"])
def test_masked_slice(tmp_path, full_water, index):
    """Test the variable shape off functionality."""
    io = znh5md.IO(tmp_path / "test.h5", variable_shape=False)
    io.extend([full_water for _ in range(10)])

    io_masked = znh5md.IO(
        tmp_path / "test.h5",
        variable_shape=False,
        mask=slice(0, 2, None),
    )
    if index == "slice":
        io_masked = io_masked[:]
        assert len(io_masked) == 10
    elif index == "list":
        io_masked = io_masked[[0, 2]]
        assert len(io_masked) == 2
    elif index == "int":
        io_masked = [io_masked[0]]
        assert len(io_masked) == 1
    elif index == "iter":
        assert len(io_masked) == 10

    for atoms in io_masked:
        assert len(atoms) == 2



import time
from dataclasses import asdict, dataclass

import numpy as np

from .abc import IOBase


@dataclass(frozen=True)
class BenchmarkMetrics:
    mean: float
    min: float
    max: float
    std: float

    def asdict(self) -> dict:
        """
        Convert the BenchmarkMetrics to a dictionary.
        """
        return asdict(self)


def benchmark_read(reader_object: IOBase, num_repeats: int = 5) -> BenchmarkMetrics:
    """
    Benchmark the read performance of a given reader object.
    Returns the mean time taken in seconds.
    """
    times = []
    for _ in range(num_repeats):
        start_time = time.time()
        _ = reader_object.read()  # Execute the read operation
        end_time = time.time()
        times.append(end_time - start_time)

    return BenchmarkMetrics(
        mean=np.mean(times), min=np.min(times), max=np.max(times), std=np.std(times)
    )



import warnings

import ase.build

import znh5md

# def test_h5md_time(tmp_path):
#     io = znh5md.IO(tmp_path / "test_time_step.h5", store="time")
#     for step in range(1, 10):
#         atoms = ase.build.molecule("H2O")
#         atoms.calc = SinglePointCalculator(atoms, energy=step * 0.1)
#         atoms.info["h5md_step"] = step
#         atoms.info["h5md_time"] = step * 0.5
#         io.append(atoms)

#     for idx, atoms in enumerate(io[:]):
#         assert atoms.info["h5md_step"] == idx + 1
#         assert atoms.info["h5md_time"] == (idx + 1) * 0.5
#         assert atoms.get_potential_energy() == (idx + 1) * 0.1

#     with h5py.File(tmp_path / "test_time_step.h5") as f:
#         npt.assert_array_equal(
#             f["particles/atoms/position/time"][:],
#             np.arange(1, 10) * 0.5
#         )
#         npt.assert_array_equal(f["particles/atoms/position/step"][:],
#                                 np.arange(1, 10))
#         npt.assert_array_equal(
#             f["observables/atoms/energy/time"][:],
#               np.arange(1, 10) * 0.5
#         )
#         npt.assert_array_equal(f["observables/atoms/energy/step"][:],
#                                np.arange(1, 10))
#         npt.assert_array_equal(
#             f["observables/atoms/energy/value"][:],
#               np.arange(1, 10) * 0.1
#         )


# def test_inconsistent_time(tmp_path):
#     images = [ase.build.molecule("H2O") for _ in range(10)]
#     images[5].info["h5md_time"] = 0.5

#     io = znh5md.IO(tmp_path / "test_inconsistent_time.h5", store="time")
#     with pytest.raises(ValueError):
#         io.extend(images)


# def test_inconsistent_step(tmp_path):
#     images = [ase.build.molecule("H2O") for _ in range(10)]
#     images[5].info["h5md_step"] = 5

#     io = znh5md.IO(tmp_path / "test_inconsistent_step.h5", store="time")
#     with pytest.raises(ValueError):
#         io.extend(images)


# def test_wrong_store(tmp_path):
#     io = znh5md.IO(tmp_path / "test_wrong_store.h5", store="linear")
#     atoms = ase.build.molecule("H2O")
#     atoms.info["h5md_step"] = 1
#     atoms.info["h5md_time"] = 0.1

#     with pytest.warns(UserWarning, match="time is ignored in 'linear' storage mode"):
#         io.append(atoms)


def test_no_warn_correct(tmp_path):
    io = znh5md.IO(tmp_path / "test_wrong_store.h5", store="linear")
    atoms = ase.build.molecule("H2O")

    # Ensure no warning is issued
    # https://docs.pytest.org/en/latest/how-to/capture-warnings.html#additional-use-cases-of-warnings-in-tests
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        io.append(atoms)



"""Test against pyh5md reference implementation."""

import ase
import numpy as np
import numpy.testing as npt
import pytest
from pyh5md import File, element

import znh5md


@pytest.fixture
def md() -> list[ase.Atoms]:
    """Run a simple MD simulation with ASE."""
    from ase import units
    from ase.calculators.emt import EMT
    from ase.calculators.singlepoint import SinglePointCalculator
    from ase.lattice.cubic import FaceCenteredCubic
    from ase.md.velocitydistribution import MaxwellBoltzmannDistribution
    from ase.md.verlet import VelocityVerlet

    size = 3

    atoms = FaceCenteredCubic(
        directions=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        symbol="Cu",
        size=(size, size, size),
        pbc=True,
    )

    atoms.calc = EMT()
    MaxwellBoltzmannDistribution(atoms, temperature_K=300)
    dyn = VelocityVerlet(atoms, 5 * units.fs)

    structures = []

    for _ in range(10):
        dyn.run(1)
        structures.append(atoms.copy())
        structures[-1].calc = SinglePointCalculator(
            atoms, energy=atoms.get_potential_energy(), forces=atoms.get_forces()
        )

    return structures


def ase_to_pyh5md(structures, path):
    with File(path, "w") as f:
        at = f.particles_group("atoms")
        at.create_box(
            dimension=3,
            boundary=["periodic"] * 3,
            store="time",
            shape=(3,),
            dtype=np.float64,
        )

        at_pos = element(
            at,
            "position",
            store="time",
            time=True,
            shape=(len(structures[0]), 3),
            dtype=np.float64,
        )

        at_species = element(
            at,
            "species",
            store="time",
            step_from=at_pos,
            time=True,
            shape=(len(structures[0]),),
            dtype=np.int32,
        )
        # TODO: test how velocities are stored / are they stored correctly?
        at_v = element(
            at,
            "velocity",
            store="time",
            shape=(len(structures[0]), 3),
            dtype=np.float64,
        )
        at_f = element(
            at, "forces", store="time", shape=(len(structures[0]), 3), dtype=np.float64
        )

        obs_at_e = element(
            f, "observables/atoms/energy", store="time", shape=(1,), dtype=np.float64
        )

        time_delta = 0.1

        for idx, atoms in enumerate(structures):
            at_pos.append(atoms.get_positions(), idx, idx * time_delta)
            at_species.append(atoms.get_atomic_numbers(), idx, idx * time_delta)
            at_v.append(atoms.get_velocities(), idx, idx * time_delta)
            obs_at_e.append(atoms.get_potential_energy(), idx * time_delta)
            at_f.append(atoms.get_forces(), idx, idx * time_delta)
            at.box.edges.append(atoms.get_cell().diagonal(), idx, idx * time_delta)


def ase_to_pyh5md_fixed_time(structures, path):
    time_delta = 0.1
    step_delta = 1

    with File(path, "w") as f:
        at = f.particles_group("atoms")

        at_pos = element(
            at,
            "position",
            store="linear",
            time=time_delta,
            step=step_delta,
            shape=(len(structures[0]), 3),
            dtype=np.float64,
        )

        at.create_box(
            dimension=3,
            boundary=["periodic"] * 3,
            store="linear",
            shape=(3,),
            dtype=np.float64,
            time=time_delta,
            step=step_delta,
        )

        at_species = element(
            at,
            "species",
            store="linear",
            time=time_delta,
            step=step_delta,
            shape=(len(structures[0]),),
            dtype=np.int32,
        )

        at_v = element(
            at,
            "velocity",
            store="linear",
            time=time_delta,
            step=step_delta,
            shape=(len(structures[0]), 3),
            dtype=np.float64,
        )
        at_f = element(
            at,
            "forces",
            store="linear",
            shape=(len(structures[0]), 3),
            dtype=np.float64,
            time=time_delta,
            step=step_delta,
        )

        obs_at_e = element(
            f,
            "observables/atoms/energy",
            store="linear",
            shape=(1,),
            dtype=np.float64,
            time=time_delta,
            step=step_delta,
        )

        for atoms in structures:
            at_pos.append(atoms.get_positions())
            at_species.append(atoms.get_atomic_numbers())
            at_v.append(atoms.get_velocities())
            obs_at_e.append(atoms.get_potential_energy())
            at_f.append(atoms.get_forces())
            at.box.edges.append(atoms.get_cell().diagonal())


@pytest.mark.parametrize("store", ["linear"])
def test_pyh5md_ASEH5MD(md, tmp_path, store):
    """Read pyh5md file with ASEH5MD."""
    path = tmp_path / "db.h5"
    if store == "time":
        ase_to_pyh5md(md, path)
    elif store == "linear":
        ase_to_pyh5md_fixed_time(md, path)
    io = znh5md.IO(path)
    structures = io[:]
    assert len(structures) == len(md)
    # check that the data is not the same
    assert md[0].get_positions()[0][0] != md[1].get_positions()[0][0]

    for a, b in zip(md, structures):
        npt.assert_array_equal(a.get_positions(), b.get_positions())
        npt.assert_array_equal(a.get_velocities(), b.get_velocities())
        npt.assert_array_equal(a.get_forces(), b.get_forces())
        npt.assert_array_equal(a.get_potential_energy(), b.get_potential_energy())
        npt.assert_array_equal(a.get_atomic_numbers(), b.get_atomic_numbers())
        npt.assert_array_equal(a.get_cell(), b.get_cell())


@pytest.mark.parametrize("store", ["linear"])
def test_DataWriter_pyh5md(md, tmp_path, store):
    """Test reading DataWriter with pyh5md."""
    path = tmp_path / "db.h5"
    io = znh5md.IO(path, timestep=0.1, store=store)
    io.extend(md)

    with File(path, "r") as f:
        g = f.particles_group("atoms")
        position = element(g, "position").value[:]
        if store == "time":
            p_time = element(g, "position").time[:]
            p_step = element(g, "position").step[:]
        elif store == "linear":
            p_time = element(g, "position").time[()]
            p_step = element(g, "position").step[()]
        species = element(g, "species").value[:]
        velocities = element(g, "velocity").value[:]
        forces = element(g, "force").value[:]
        energy = element(f, "observables/atoms/potential_energy").value[:]
        cell = element(g, "box/edges").value[:]

    assert len(position) == len(md)
    for idx, atoms in enumerate(md):
        npt.assert_array_equal(position[idx], atoms.get_positions())
        npt.assert_array_equal(species[idx], atoms.get_atomic_numbers())
        npt.assert_array_equal(velocities[idx], atoms.get_velocities())
        npt.assert_array_equal(forces[idx], atoms.get_forces())
        npt.assert_array_equal(energy[idx], atoms.get_potential_energy())
        npt.assert_array_equal(cell[idx], atoms.get_cell())
        if store == "time":
            assert p_time[idx] == idx * 0.1
            assert p_step[idx] == idx
        elif store == "linear":
            assert p_time == 0.1
            assert p_step == 1
